# -*- coding: utf-8 -*-

"""new version of ddf_reader, base on datapackage.json"""

import os
import pandas as pd
from . import config
from .datapackage import get_datapackage
from pprint import pformat


# main class for ddf reading
class DDF():
    """DDF reader class

    The reader instance accepts an dataset id or absolute path on init. If absolute path is given,
    it will load the dataset in path. Other wise it will search the dataset id in the path set by
    the `DDF_SEARCH_PATH` global variable.
    """
    def __init__(self, ddf_id, no_check_valid=False):
        if os.path.isabs(ddf_id):
            self.dataset_path = ddf_id
            self.ddf_id = os.path.dirname(ddf_id)
        else:
            self.dataset_path = os.path.join(config.DDF_SEARCH_PATH, ddf_id)
            self.ddf_id = ddf_id
        if not no_check_valid:
            assert is_dataset(self.dataset_path), \
                "path is not ddf dataset: {}".format(self.dataset_path)
        self._datapackage = None
        self._concepts = None

    def __repr__(self):
        return "dataset {}".format(self.ddf_id)

    def describe(self):
        return ("concepts in this dataset: \n\n{}"
                .format(pformat(self.concepts['concept_type'].to_dict())))

    @property
    def datapackage(self):
        """the datapackage object. create one if it doesn't exist"""
        if not self._datapackage:
            self._datapackage = get_datapackage(self.dataset_path, update=False)
        return self._datapackage

    @property
    def concepts(self):
        """convenient function to get all concepts"""
        if self._concepts is None:
            self._concepts = self.get_concepts()
        return self._concepts

    @property
    def dtypes(self):
        """return a mapping for column -> python object type. internal use only."""
        concept_types = self.concepts['concept_type']
        res = dict()
        for c, v in concept_types.iteritems():
            if v in ['string', 'entity_domain', 'entity_set']:
                res[c] = str
            elif v in ['time']:
                res[c] = int  # TODO: support more time format
            else:  # for measures and other concept type, let pandas decide
                continue

        return res

    @property
    def indicator_dict(self):
        """return all indicators"""
        return dict([name, list(item.keys())]
                    for name, item in self.get_datapoint_files().items())

    def get_all_files(self):
        """return a list of all files in this dataset"""
        resources = self.datapackage['resources']
        return [x['path'] for x in resources]

    def get_concept_files(self):
        """return a list of all concept files"""
        resources = self.datapackage['resources']
        return [x['path'] for x in resources if x['schema']['primaryKey'] == 'concept']

    def get_concepts(self, concept_type='all', **kwargs):
        """get concepts from the concept table

        Args
        ----
        concept_type : `str` or `list`
            the concept type(s) to filter, if 'all' is provided, all concepts will be returned
        **kwargs : dict
            keyword arguments provided to :py:func:`pandas.read_csv`

        Returns
        -------
        DataFrame
            concept table

        """
        concept_files = self.get_concept_files()
        all_concepts = pd.concat([
                pd.read_csv(os.path.join(self.dataset_path, x),
                            index_col='concept', **kwargs) for x in concept_files])
        if concept_type == 'all':
            return all_concepts
        elif isinstance(concept_type, str):
            return all_concepts[all_concepts.concept_type == concept_type]
        elif isinstance(concept_type, list):
            return all_concepts[all_concepts.concept_type.isin(concept_type)]

    def get_entities(self, domain=None, **kwargs):
        """get entities from one or all domains

        Args
        ----
        domain : `str`, optional
            the domain to filter, if not provided, all domain will be returned
        **kwargs : dict
            keyword arguments provided to :py:func:`pandas.read_csv`

        Returns
        -------
        dict
            a dictionary like ``{entity_set_name : DataFrame}``

        """
        resources = self.datapackage['resources']
        entity_concepts = self.get_concepts(['entity_domain', 'entity_set'])

        if domain:
            entity_concepts = entity_concepts.reset_index()
            if 'domain' in entity_concepts.columns:
                mask = ((entity_concepts.domain == domain) |
                        ((entity_concepts.concept == domain) &
                         (entity_concepts.concept_type == 'entity_domain')))
            else:
                mask = entity_concepts.concept == domain
            entity_concepts = entity_concepts[mask].set_index('concept')

        if 'dtype' in kwargs.keys():
            dtype = kwargs.pop('dtype')
        else:
            dtype = self.dtypes

        entities = dict()
        for res in resources:
            key = res['schema']['primaryKey']
            if isinstance(key, str) and key != 'concept':
                name = res['name'].split('--')[-1]  # TODO: don't judge by name
                if (res['schema']['primaryKey'] in list(entity_concepts.index) or
                    (domain is not None and res['schema']['primaryKey'] == domain)):
                    entities[name] = pd.read_csv(
                        os.path.join(self.dataset_path, res['path']),
                        dtype=dtype, **kwargs)
                    entities[name] = entities[name].set_index(res['schema']['primaryKey'])
        return entities

    def get_datapoint_files(self):
        """return a list of datapoints files"""
        datapoints = dict()
        resources = self.datapackage['resources']

        for res in resources:
            key = res['schema']['primaryKey']
            if isinstance(key, list):
                key = tuple([x for x in key])
                fields = [x['name'] for x in res['schema']['fields']]
                name = [x for x in fields if x not in key]
                assert len(name) == 1
                name = name[0]
                if name in datapoints.keys():
                    if key in datapoints[name].keys():
                        datapoints[name][key].append(res['path'])
                    else:
                        datapoints[name][key] = [res['path']]
                else:
                    datapoints[name] = {key: [res['path']]}
        return datapoints

    def __build_datapoint_df(self, files):
        return pd.concat(
            [pd.read_csv(os.path.join(self.dataset_path, x), dtype=self.dtypes) for x in files],
            ignore_index=True
        )

    def get_datapoints(self, measure=None, primaryKey=None):
        """get datapoints, filter by concept or primaryKey

        Args
        ----
        measure : `str`, optional
            only get this measure
        primaryKey : `str`, optional
            only get this primaryKey

        Returns
        -------
        dict
            A dictionary like ``{ measure_name : { primaryKeys : DataFrame }}``
        """
        datapoint_files = self.get_datapoint_files()
        datapoints = dict()

        if measure:
            if primaryKey:  # both measure and primaryKey given
                datapoints[measure] = {
                    primaryKey: (self.__build_datapoint_df(datapoint_files[measure][primaryKey])
                                 .set_index(list(primaryKey)))
                }
            else:  # only measure given
                datapoints[measure] = dict([
                        (k, self.__build_datapoint_df(
                                datapoint_files[measure][k]).set_index(list(k)))
                        for k in datapoint_files[measure].keys()])
        else:
            if primaryKey:  # only primaryKey given
                for m in datapoint_files.keys():
                    datapoints[m] = {
                        primaryKey: (self.__build_datapoint_df(datapoint_files[m][primaryKey])
                                     .set_index(list(primaryKey)))
                    }
            else:  # no parameters, return all
                for m in datapoint_files.keys():
                    datapoints[m] = dict([
                            (k, self.__build_datapoint_df(
                                    datapoint_files[m][k]).set_index(list(k)))
                            for k in datapoint_files[m].keys()])
        return datapoints

    def get_datapoint_df(self, measure, primaryKey=None):
        """get datapoints by measure, returns a DataFrame

        Args
        ----
        measure : `str`
            the measure to get
        primaryKey : `str`, optional
            the primaryKey to get, if not provided and the datapoint have multiple
            primaryKeys avaliable, the first primaryKey in datapackage will be returned

        """
        datapoint_files = self.get_datapoint_files()

        if len(datapoint_files[measure].keys()) == 1:
            keys = list(datapoint_files[measure].keys())[0]
            if primaryKey:
                if not set(keys) == set(primaryKey):
                    raise KeyError('no such key for the measure: ', primaryKey)
            df = self.get_datapoints(measure, keys)
            return df[measure][keys]
        else:
            if not primaryKey:
                raise ValueError("please specify a primaryKey for measures with multiple "
                                 "primaryKeys")
            for keys in datapoint_files[measure].keys():
                if set(keys) == set(primaryKey):
                    df = self.get_datapoints(measure, keys)
                    return df[measure][keys]
            raise ValueError('key not found for the measure!')


# helper functions:
# check if a directory is dataset root dir
def is_dataset(path):
    """check if a directory is a dataset directory

    This function checks if ddf--index.csv and datapackage.json exists
    to judge if the dir is a dataset.
    """
    index_path = os.path.join(path, 'ddf--index.csv')
    datapackage_path = os.path.join(path, 'datapackage.json')
    if os.path.exists(index_path) or os.path.exists(datapackage_path):
        return True
    else:
        return False


# function for listing all ddf projects
def list_datasets():
    """list all availabile datasets"""
    datasets = []
    for d in next(os.walk(config.DDF_SEARCH_PATH))[1]:
        dataset_path = os.path.join(config.DDF_SEARCH_PATH, d)
        if is_dataset(dataset_path):
            datasets.append(d)
    return datasets



# -*- coding: utf-8 -*-

"""window procedure for recipes"""

import logging
from typing import List

from .. helpers import debuggable, read_opt, mkfunc
from .. model.ingredient import DataPointIngredient
from .. model.chef import Chef


logger = logging.getLogger('window')


@debuggable
def window(chef: Chef, ingredients: List[DataPointIngredient], result, **options) -> DataPointIngredient:
    """apply functions on a rolling window

    .. highlight:: yaml

    Procedure format:

    ::

       procedure: window
       ingredients:  # list of ingredient id
         - ingredient_id
       result: str  # new ingredient id
       options:
         window:
           column: str  # column which window is created from
           size: int or 'expanding'  # if int then rolling window, if expanding then expanding window
           min_periods: int  # as in pandas
           center: bool  # as in pandas
           aggregate: dict

    Two styles of function block are supported, and they can mix in one procedure:

    ::

       aggregate:
         col1: sum  # run rolling sum to col1
         col2: mean  # run rolling mean to col2
         col3:  # run foo to col3 with param1=baz
           function: foo
           param1: baz

    Keyword Args
    ------------
    window: dict
        window definition, see above for the dictionary format
    aggregate: dict
        aggregation functions

    Examples
    --------

    An example of rolling windows:

    .. highlight:: yaml

    ::

        procedure: window
        ingredients:
            - ingredient_to_roll
        result: new_ingredient_id
        options:
          window:
            column: year
            size: 10
            min_periods: 1
            center: false
          aggregate:
            column_to_aggregate: sum

    Notes
    -----
    Any column not mentioned in the `aggregate` block will be dropped in the returned ingredient.
    """
    assert len(ingredients) == 1, "procedure only support 1 ingredient for now."
    # ingredient = chef.dag.get_node(ingredients[0]).evaluate()
    ingredient = ingredients[0]
    logger.info('window: ' + ingredient.id)

    # reading options
    window = options.pop('window')
    aggregate = options.pop('aggregate')

    column = read_opt(window, 'column', required=True)
    size = read_opt(window, 'size', required=True)
    min_periods = read_opt(window, 'min_periods', default=0)
    center = read_opt(window, 'center', default=False)

    data = ingredient.compute()
    newdata = dict()

    for k, func in aggregate.items():
        f = mkfunc(func)
        # keys for grouping. in multidimensional data like datapoints, we want create
        # groups before rolling. Just group all key column except the column to aggregate.
        keys = ingredient.key.copy()
        keys.remove(column)
        df = data[k].set_index(ingredient.key)
        levels = [df.index.names.index(x) for x in keys]
        if size == 'expanding':
            newdata[k] = (df.groupby(level=levels, group_keys=False)
                          .expanding(on=column, min_periods=min_periods, center=center)
                          .agg(f).reset_index().dropna())
        else:
            # There is a bug when running rolling on with groupby in pandas.
            # see https://github.com/pandas-dev/pandas/issues/13966
            # We will implement this later when we found work around or it's fixed
            # for now, we don't use the `on` parameter in rolling.
            # FIXME: add back the `on` parameter.
            newdata[k] = (df.groupby(level=levels, group_keys=False)
                          .rolling(window=size, min_periods=min_periods, center=center)
                          .agg(f).reset_index().dropna())

    return DataPointIngredient.from_procedure_result(result, ingredient.key, newdata)



# -*- coding: utf-8 -*-

"""commonly used calculation methods"""

import pandas as pd
import numpy as np


# Groupby Aggrgate


# Groupby Transform
def zcore(x):
    return (x - x.mean()) / x.std()


# Groupby Filter
def gt(x, val, how='all', include_eq=False):
    f = getattr(np, how)
    if include_eq:
        return f(x >= val)
    else:
        return f(x > val)


def lt(x, val, how='all', include_eq=False):
    f = getattr(np, how)
    if include_eq:
        return f(x <= val)
    else:
        return f(x <= val)


def between(x, upper, lower, how='all', include_upper=False, include_lower=False):
    return gt(x, lower, how, include_lower) and lt(x, upper, how, include_upper)


# Rolling
def aagr(df: pd.DataFrame, window: int=10):  # TODO: don't include the window
    """average annual growth rate

    Parameters
    ----------
    window : `int`
        the rolling window size

    Returns
    -------
    return : `DataFrame`
        The rolling apply result
    """
    pct = df.pct_change()
    return pct.rolling(window).apply(np.mean).dropna()



# -*- coding: utf-8 -*-

import os
import sys
import hashlib
import logging
import json
from functools import wraps, partial
from collections.abc import Sequence, Mapping
from time import time

import click
import numpy as np
import pandas as pd
import dask.dataframe as dd

from joblib import Memory
from tempfile import mkdtemp

from . import ops
from ..model.package import DDFcsv


memory = Memory(location=mkdtemp(), verbose=0)


@memory.cache
def read_local_ddf(ddf_id, base_dir='./'):
    """read local dataset and return the DDF object."""
    if os.path.isabs(ddf_id):
        return DDFcsv.from_path(ddf_id).ddf
    else:
        path = os.path.join(base_dir, ddf_id)
        return DDFcsv.from_path(path).ddf


def create_dsk(data, parts=10):
    """given a dictionary of {string: pandas dataframe}, create a new dictionary with dask dataframe"""
    # TODO: check the best parts to use
    for k, v in data.items():
        if isinstance(v, pd.DataFrame):
            data[k] = dd.from_pandas(v, npartitions=parts)
    return data


def dsk_to_pandas(data):
    """The reverse for create_dsk function"""
    for k, v in data.items():
        if isinstance(v, dd.DataFrame):
            data[k] = v.compute()
    return data


def build_dictionary(chef, dict_def, ignore_case=False):
    """build a dictionary from a dictionary definition"""
    if (len(dict_def) == 3 and
            'base' in dict_def and
            'key' in dict_def and
            'value' in dict_def):
        value = dict_def['value']
        key = dict_def['key']
        if isinstance(key, str):
            keys = [key]
        else:
            keys = key
        ingredient = chef.dag.node_dict[dict_def['base']].evaluate()
        if ingredient.dtype == 'synonyms':
            di = ingredient.get_data()[value]   # synonyms data is already dict
            if ignore_case:
                res = dict()
                for k, v in di.items():
                    res[k.lower()] = v
            else:
                res = di.copy()
            return res
        elif ingredient.dtype == 'entities':
            df = ingredient.get_data()[ingredient.key]
            return build_dictionary_from_dataframe(df, keys, value, ignore_case)
        else:
            raise NotImplementedError('unsupported data type {}'.format(ingredient.dtype))
    elif isinstance(dict_def, str):
        base_path = chef.config['dictionaries_dir']
        path = os.path.join(base_path, dict_def)
        return build_dictionary_from_file(path)
    else:
        return dict_def


def build_dictionary_from_file(file_path):
    d = json.load(open(file_path, 'r'))
    assert isinstance(d, dict)
    return d


def build_dictionary_from_dataframe(df, keys, value, ignore_case=False):
    dic = dict()
    for k in keys:
        d = (df[[k, value]]
             .dropna(how='any')
             .set_index(k)[value]
             .to_dict())
        for i, v in d.items():
            if ignore_case:
                if isinstance(i, str):
                    i = i.lower()
            if i not in dic.keys():
                dic[i] = v
            elif dic[i] != v:
                raise KeyError("ambiguous key: {} is mapped "
                               "to both {} and {}".format(i, dic[i], v))
            else:
                continue
    return dic


def prompt_select(selects, text_before=None):
    """ask user to choose in a list of options"""
    def value_proc(v):
        if v == 'n':
            return -1
        if v == 'q':
            import sys
            sys.exit(130)  # code for user interrupted
        return int(v)

    if text_before:
        click.echo(text_before)
    select_text = []

    for i, v in enumerate(selects):
        select_text.append('{}: {}'.format(i+1, v))
    select_text = '\n'.join(select_text)
    click.echo(select_text)

    prompt_text = 'Please select a value ({} ~ {}), or "n" to skip, "q" to quit'\
                  .format(1, len(selects))
    val = click.prompt(prompt_text, value_proc=value_proc)
    if val == -1:
        return -1
    return selects[val-1]


def sort_df(df, key, sort_key_columns=True, custom_column_order=None):
    """Sorting df columns and rows.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to sort
    key : `str` or list
        columns of dataframe, to be used as sorting key(s)

    Keyword Args
    ------------
    sort_key_columns : bool
        whehter to sort index column orders. If false index columns will retain the order of `key` parameter.
    custom_column_order : dict
        column weights for columns except keys. Columns not mentioned will have 0 weight.
        Bigger weight means higher rank.
    """
    if isinstance(key, str):
        key = [key]

    cols_minus_key = df.columns.tolist()
    for k in key:
        cols_minus_key.remove(k)

    if custom_column_order:
        order = dict([(x, 0) for x in cols_minus_key])
        order.update(custom_column_order)
        # sort in a dataframe and back to list.
        order_new = (pd.DataFrame.from_dict(order, orient='index', columns=['Ord'])
                     .sort_values(by='Ord', ascending=False).index.tolist())
        cols_minus_key_new = [x for x in order_new if x in cols_minus_key]
        cols_minus_key = cols_minus_key_new
    else:
        cols_minus_key.sort()

    if sort_key_columns:
        key.sort()

    cols_new = [*key, *cols_minus_key]

    # change categorical type to string.
    # we need to do this because we didn't define order for category
    # when we read the data
    cat_cols = df.select_dtypes(include=['category']).columns.tolist()
    for c in cat_cols:
        df[c] = df[c].astype('str')

    df = df.sort_values(by=key)

    return df[cols_new]


def read_opt(options, key, required=False, default=None, method='get'):
    """utility to read an attribute from an options dictionary

    Parameters
    ----------
    options : dict
        the option dictionary to read
    key : `str`
        the key to read

    Keyword Args
    ------------
    required : bool
        if true, raise error if the `key` is not in the option dict
    default : object
        a default to return if `key` is not in option dict and `required` is false
    """
    if key in options.keys():
        if method == 'get':
            return options.get(key)
        elif method == 'pop':
            return options.pop(key)
        else:
            raise ValueError("{} is not supported method".format(method))
    if required:
        raise KeyError('Field "{}" is mandatory. Please provide this field in the options.'.format(key))
    return default


def mkfunc(options):
    """create function warppers base on the options provided

    This function is used in procedures which have a function block. Such as
    :py:func:`ddf_utils.chef.procedure.groupby`. It will try to return functions
    from numpy or :py:mod:`ddf_utils.ops`.

    Parameters
    ----------
    options : `str` or dict
        if a dictionary provided, "function" should be a key in the dictionary
    """
    if isinstance(options, str):
        return getattr(np, options)
    else:
        func = getattr(ops, options.pop('function'))
        return partial(func, **options)


def get_procedure(procedure, base_dir):
    """return a procedure function from the procedure name

    Parameters
    ----------
    procedure : `str`
        the procedure to get, supported formats are
        1. procedure: sub/dir/module.function
        2. procedure: module.function
    base_dir : `str`
        the path for searching procedures
    """
    import ddf_utils.chef.procedure as pc
    if '.' in procedure:
        assert 'base_dir' is not None, "please set procedure_dir in config if you have custom procedures"
        sys.path.insert(0, base_dir)
        module_name, func_name = procedure.split('.')
        _mod = __import__(module_name)
        func = getattr(_mod, func_name)
        sys.path.remove(base_dir)
    else:
        func = getattr(pc, procedure)

    return func


def gen_sym(key, others=None, options=None):
    """generate symbol for chef ingredient/procedure result"""
    tail = hashlib.sha256((str(others) + str(options) + str(time())).encode('utf8')).hexdigest()[:6]
    return '{}_{}'.format(key, tail)


def gen_query(conds, scope=None, available_scopes=None):
    """generate dataframe query from mongo-like queries"""
    comparison_keywords = {
        '$eq': '==',
        '$gt': '>',
        '$gte': '>=',
        '$lt': '<',
        '$lte': '>',
        '$ne': '!=',
        '$in': 'in',
        '$nin': 'not in'
    }
    logical_keywords = ['$and', '$or', '$not', '$nor']

    def process_values(n):
        if isinstance(n, str):
            return "'{}'".format(n)
        else:
            return str(n)

    res = []

    for key, val in conds.items():
        if key in logical_keywords:
            if key in ['$and', '$or']:
                to_join = ' {} '.format(key[1:])
                queries = [gen_query({k: v}, scope, available_scopes) for k, v in val.items()]
                return '({})'.format(to_join.join(filter(lambda x: x != '', queries)))
            if key == '$not':
                queries = [gen_query({k: v}, scope, available_scopes) for k, v in val.items()]
                return '~({})'.format(' and '.join(filter(lambda x: x != '', queries)))
            if key == '$nor':
                queries = [gen_query({k: v}, scope, available_scopes) for k, v in val.items()]
                return ' and '.join(['~({})'.format(filter(lambda x: x != '', queries))])
        else:
            if key in comparison_keywords.keys():
                assert not isinstance(val, Mapping)
                assert scope != None
                res.append("{} {} {}".format(scope,
                                             comparison_keywords[key],
                                             process_values(val)))
            elif isinstance(val, str):
                res.append("{} == {}".format(key, process_values(val)))
            elif isinstance(val, Sequence):
                res.append("{} in {}".format(key, process_values(val)))
            else:
                if available_scopes is None or key in available_scopes:
                    res.append(gen_query(val, scope=key, available_scopes=available_scopes))
    return ' and '.join(res)


def query(df, conditions, available_scopes=None):
    """query a dataframe with mongo-like queries"""
    q = gen_query(conditions, available_scopes=available_scopes)
    logging.info("querying: {}".format(q))
    if q == '' or q == '()':
        logging.warning("empty query")
        return df
    return df.query(q)


# def procedure(func, **options_dict):
#     """The procedure wrapper, to create a procedure

#     @procedure
#     def procedure_func(**options):
#         pass

#     will automatically create this:

#     def procedure_func(chef=chef, ingredients=[], result='result', **options)

#     """
#     @wraps(func)
#     @debuggable
#     def procedure_func(chef, ingredients, result, *args, **kwargs):
#         return
#     return procedure_func


def debuggable(func):
    """return a function that accepts `debug` as keyword parameters."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        if 'debug' in kwargs.keys():
            debug = kwargs.pop('debug')
        else:
            debug = False

        if 'breakpoint' in kwargs.keys():
            bk = kwargs.pop('breakpoint')
        else:
            bk = False

        if bk:
            import ipdb
            result = ipdb.runcall(func, *args, **kwargs)
            ipdb.pm()
        else:
            result = func(*args, **kwargs)

        if debug:
            chef = args[0]
            debug_path = chef.config.get('debug_output_path', None)
            if debug_path is None:
                logging.warning('debug output path not set!')
                chef.add_config(debug_output_path='./_debug')
                debug_path = './_debug'
            outpath = os.path.join(debug_path, result.id)
            if os.path.exists(outpath):
                import shutil
                shutil.rmtree(outpath)
                os.mkdir(outpath)
            result.serve(outpath)
        return result
    return wrapper



# -*- coding: utf-8 -*-

"""datapackage model"""

import os.path as osp
from typing import List, Tuple, Dict, Union, Callable
import attr
import json
from itertools import product
from collections import OrderedDict
from tqdm import tqdm

import pandas as pd

from .ddf import DDF, Concept, EntityDomain, Entity, DaskDataPoint, Synonym
from .utils import absolute_path

import logging

logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, repr=False)
class TableSchema:
    """Table Schema Object Class"""
    fields: List[dict]
    primaryKey: Union[List[str], str]

    @classmethod
    def from_dict(cls, d: dict):
        fields = d['fields']
        primaryKey = d['primaryKey']
        return cls(fields, primaryKey)

    @property
    def field_names(self):
        return [f['name'] for f in self.fields]

    @property
    def common_fields(self):
        field_names = self.field_names
        pkey = self.primaryKey
        if isinstance(pkey, str):
            common_fields = list(filter(lambda x: x != pkey, field_names))
        else:
            common_fields = list(filter(lambda x: x not in pkey, field_names))
        return common_fields

    def __repr__(self):
        return "TableSchema(primaryKey: {}, fields: {})".format(self.primaryKey, self.common_fields)


@attr.s(auto_attribs=True)
class Resource:
    name: str
    path: str
    schema: TableSchema

    @classmethod
    def from_dict(cls, d: dict):
        path = d['path']
        name = d['name']
        schema = TableSchema.from_dict(d['schema'])
        return cls(name, path, schema)

    def to_dict(self):
        res = vars(self).copy()
        if 'schema' in res:
            res['schema'] = vars(res['schema']).copy()
        return res


@attr.s(auto_attribs=True)
class DDFSchema:
    primaryKey: List[str]
    value: str
    resources: List[str]  # a list of resource names

    @classmethod
    def from_dict(cls, d: dict):
        primaryKey = d['primaryKey']
        value = d['value']
        resources = d['resources']
        return cls(primaryKey=primaryKey, value=value, resources=resources)


@attr.s(auto_attribs=True, repr=False)
class DataPackage:
    base_path: str
    resources: List[Resource]
    props: dict = attr.ib(factory=dict)

    def __attrs_post_init__(self):
        self.base_path = absolute_path(self.base_path)

    def __repr__(self):
        return f"DataPackage({self.base_path})"

    @classmethod
    def from_dict(cls, d_: dict, base_path='./'):
        d = d_.copy()
        resources = list(map(Resource.from_dict, d.pop('resources')))
        return cls(base_path=base_path, resources=resources, props=d)

    @classmethod
    def from_json(cls, json_path):
        # TODO: security checking.
        json_path = absolute_path(json_path)
        base_path = osp.dirname(json_path)
        d = json.load(open(json_path))
        return cls.from_dict(d, base_path)

    @classmethod
    def from_path(cls, path):
        path = absolute_path(path)
        json_path = osp.join(path, 'datapackage.json')
        return cls.from_json(json_path)

    def to_dict(self):
        """dump the datapackage to disk"""
        raise NotImplementedError


@attr.s(repr=False)
class DDFcsv(DataPackage):
    """DDFCSV datapackage."""
    ddfSchema: Dict[str, List[DDFSchema]] = attr.ib(factory=dict)
    ddf: DDF = attr.ib(init=False)
    concepts_resources: List[Resource] = attr.ib(init=False)
    entities_resources: List[Resource] = attr.ib(init=False)
    datapoints_resources: List[Resource] = attr.ib(init=False)
    synonyms_resources: List[Resource] = attr.ib(init=False)

    # config for read_csv
    _default_reader_options = {'keep_default_na': False, 'na_values': ['']}

    def __attrs_post_init__(self):
        super(DDFcsv, self).__attrs_post_init__()
        conc = list()
        ent = list()
        dp = list()
        syn = list()
        for r in self.resources:
            pkey = r.schema.primaryKey
            if isinstance(pkey, str):
                if pkey == 'concept':
                    conc.append(r)
                else:
                    ent.append(r)
            else:  # TODO: datapoints key might be one column, not list of columns?
                if 'synonym' in pkey:
                    syn.append(r)
                else:
                    dp.append(r)
        self.concepts_resources = conc
        self.entities_resources = ent
        self.datapoints_resources = dp
        self.synonyms_resources = syn
        self.ddf = self.load_ddf()

    @classmethod
    def from_dict(cls, d_: dict, base_path='./'):
        d = d_.copy()
        resources = list(map(Resource.from_dict, d.pop('resources')))
        if 'ddfSchema' in d.keys():
            ddf_schema_ = d.pop('ddfSchema')
            ddf_schema = dict()
            for k, v in ddf_schema_.items():
                ddf_schema[k] = [DDFSchema.from_dict(d) for d in v]
        else:
            ddf_schema = {}
        return cls(base_path=base_path, resources=resources, ddfSchema=ddf_schema, props=d)

    def to_dict(self):
        res = OrderedDict(self.props.copy())
        res['resources'] = [r.to_dict() for r in self.resources]
        if self.ddfSchema:
            res['ddfSchema'] = dict()
            for k, v in self.ddfSchema.items():
                res['ddfSchema'][k] = [vars(sch).copy() for sch in v]
        return res

    def _gen_concepts(self):
        concepts_paths = [osp.join(self.base_path, r.path) for r in self.concepts_resources]
        for p in concepts_paths:
            df = pd.read_csv(p, index_col='concept', dtype=str, **self._default_reader_options)
            for concept, row in df.iterrows():
                concept_type = row['concept_type']
                props = row.drop('concept_type').to_dict()
                yield (concept, Concept(id=concept, concept_type=concept_type, props=props))

    def _gen_entities(self, concepts: Dict[str, Concept]):
        for r in self.entities_resources:
            pkey = r.schema.primaryKey
            if concepts[pkey].concept_type == 'entity_domain':
                domain = concepts[pkey].id
            else:
                domain = concepts[pkey].props['domain']

            df = pd.read_csv(osp.join(self.base_path, r.path), dtype=str,  # TODO: is it okay to use str for all?
                             **self._default_reader_options)
            df = df.set_index(pkey)
            is_cols = list(filter(lambda x: x.startswith('is--'), df.columns.values))
            for ent, row in df.iterrows():
                sets = list()
                for c in is_cols:
                    if row[c] == 'TRUE' and c[4:] != domain:
                        sets.append(c[4:])  # strip the 'is--' part, only keep set name
                yield (domain, Entity(id=ent, domain=domain, sets=sets, props=row.drop(is_cols).to_dict()))

    def _gen_datapoints(self):
        for r in self.datapoints_resources:
            fields = r.schema.common_fields
            pkey = r.schema.primaryKey
            for f in fields:
                yield (f, pkey, osp.join(self.base_path, r.path))

    def _gen_synonyms(self):
        for r in self.synonyms_resources:
            # there should be only two columns
            pkey = r.schema.primaryKey
            if pkey[0] == 'synonym':
                concept = pkey[1]
            else:
                concept = pkey[0]
            df = pd.read_csv(osp.join(self.base_path, r.path), **self._default_reader_options)
            sym = Synonym(concept_id=concept, synonyms=df.set_index('synonym')[concept].to_dict())
            yield (concept, sym)

    @staticmethod
    def entity_domain_to_categorical(domain: EntityDomain):
        entities = [e.id for e in domain.entities]
        return pd.api.types.CategoricalDtype(entities)

    @staticmethod
    def entity_set_to_categorical(domain: EntityDomain, s: str):
        entity_set = domain.get_entity_set(s)
        entities = [e.id for e in entity_set]
        return pd.api.types.CategoricalDtype(entities)

    def load_ddf(self):
        """-> DDF"""
        # load concepts
        concepts = dict(self._gen_concepts())

        # load entities
        entities = list(self._gen_entities(concepts))
        domains = dict()
        for domain, entity in entities:
            if domain not in domains.keys():
                domains[domain] = EntityDomain(id=domain, entities=[])

            domains[domain].add_entity(entity)

        # load datapoints. Here we will use Dask for all
        # 1. create categories for entity domains
        dtypes = dict()
        for domain_name, domain in domains.items():
            dtypes[domain_name] = self.entity_domain_to_categorical(domain)
            for eset in domain.entity_sets:
                dtypes[eset] = self.entity_set_to_categorical(domain, eset)
        # 2. dtype for time
        for c_id, c in concepts.items():
            if c.concept_type == 'time':
                dtypes[c_id] = 'int16'  # TODO: maybe there are other format for time.
        # 3. group files for same indicator together
        indicators = dict()
        for field, pkey, path in self._gen_datapoints():
            # import ipdb; ipdb.set_trace()
            indicator = field
            pkey = tuple(sorted(pkey))
            if indicator not in indicators:
                indicators.setdefault(indicator, dict([(pkey, [path])]))
            else:
                if pkey not in indicators[indicator]:
                    indicators[indicator][pkey] = [path]
                else:
                    indicators[indicator][pkey].append(path)
        datapoints = dict()
        for i, v in indicators.items():
            datapoints[i] = dict()
            dtypes_ = dtypes.copy()
            # dtypes_[i] = 'float'  # TODO: supporting string/float datatypes, not just float
            for k, paths in v.items():
                dp = DaskDataPoint(id=i, dimensions=k, path=paths, dtypes=dtypes_)
                datapoints[i][k] = dp

        # load synonyms
        synonyms = dict(self._gen_synonyms())

        # return complete DDF object
        return DDF(concepts=concepts, entities=domains, datapoints=datapoints, synonyms=synonyms, props=self.props)

    def generate_ddf_schema(self):
        hash_table = {}
        ddf_schema = {'concepts': [], 'entities': [], 'datapoints': [], 'synonyms': []}
        entity_value_cache = dict()
        dtypes = dict()

        # generate set-membership details for every single entity in dataset
        # also create dtypes for later use
        for domain_id, domain in self.ddf.entities.items():
            dtypes[domain_id] = self.entity_domain_to_categorical(domain)
            entity_value_cache[domain_id] = dict()
            for ent in domain.entities:
                sets = set()
                sets.add(domain_id)
                for s in ent.sets:
                    sets.add(s)
                entity_value_cache[domain_id][ent.id] = tuple(sets)

        def _which_sets(entity_, domain_):
            try:
                # FIXME: make sure entity id are just str, earlier
                e = str(entity_)
                return entity_value_cache[domain_][e]
            except KeyError:
                # possible cause is that the domain does not contain enough entities.
                # TODO: give warning when creating entity domain
                logger.debug('entity {} is not in {} domain!'.format(entity_, domain_))
                raise

        def _gen_key_value_object(resource: Resource):
            logger.debug('working on: {}'.format(resource.path))
            if isinstance(resource.schema.primaryKey, str):
                pkeys = [resource.schema.primaryKey]
            else:
                pkeys = resource.schema.primaryKey

            entity_cols = [x for x in pkeys
                           if x in self.ddf.concepts
                           and self.ddf.concepts[x].concept_type in ['entity_domain', 'entity_set']]
            value_cols = resource.schema.common_fields
            data = pd.read_csv(osp.join(self.base_path, resource.path), dtype=dtypes,
                               **self._default_reader_options)

            # for resources that have entity_columns: only consider all permutations on entity columns
            if len(entity_cols) > 0:
                data = data[entity_cols].drop_duplicates()

            pkeys_prop = dict()
            for c in pkeys:
                if c not in self.ddf.concepts:
                    pkeys_prop[c] = {'type': 'non_concept'}
                else:
                    concept = self.ddf.concepts[c]
                    if concept.concept_type == 'entity_set':
                        pkeys_prop[c] = {'type': 'entity_set',
                                         'domain': concept.props['domain']}
                    elif concept.concept_type == 'entity_domain':
                        pkeys_prop[c] = {'type': 'entity_domain'}
                    else:
                        pkeys_prop[c] = {'type': 'others'}

            all_permutations = set()
            for _, r in data.iterrows():
                perm = list()
                for c in pkeys:
                    if pkeys_prop[c]['type'] == 'entity_set':
                        domain = pkeys_prop[c]['domain']
                        perm.append(_which_sets(r[c], domain))
                    elif pkeys_prop[c]['type'] == 'entity_domain':
                        perm.append(_which_sets(r[c], c))
                    else:
                        perm.append(tuple([c]))

                all_permutations.add(tuple(perm))

            for row in all_permutations:
                for perm in product(*row):
                    if len(value_cols) > 0:
                        for c in value_cols:
                            obj = {'primaryKey': list(perm), 'value': c, 'resource': resource.name}
                            logger.debug('yielding: {}'.format(str(obj)))
                            yield obj
                    else:
                        obj = {'primaryKey': list(perm), 'value': None, 'resource': resource.name}
                        logger.debug('yielding: {}'.format(str(obj)))
                        yield obj

        def _add_to_schema(resource_schema):
            """handle objects generated by ``_gen_key_value_object``"""
            key = '-'.join(sorted(resource_schema['primaryKey']))
            if not pd.isnull(resource_schema['value']):
                hash_val = key + '--' + resource_schema['value']
            else:
                hash_val = key + '--' + 'nan'
            if hash_val not in hash_table.keys():
                hash_table[hash_val] = {
                    'primaryKey': sorted(resource_schema['primaryKey']),
                    'value': resource_schema['value'],
                    'resources': {resource_schema['resource']}
                }
            else:
                hash_table[hash_val]['resources'].add(resource_schema['resource'])

        # make progressbar and run the process to generate schema
        # TODO: improve progress bar codes

        if logger.getEffectiveLevel() != 10:
            pbar = tqdm(total=len(self.resources))

        for g in map(_gen_key_value_object, self.resources):
            if logger.getEffectiveLevel() != 10:
                pbar.update(1)
            for kvo in g:
                logging.debug("adding kvo {}".format(str(kvo)))
                _add_to_schema(kvo)

        if logger.getEffectiveLevel() != 10:
            pbar.close()

        for sch in hash_table.values():
            sch['resources'] = list(sch['resources'])  # convert set to list
            sch_object = DDFSchema.from_dict(sch)
            if len(sch['primaryKey']) == 1:
                if sch['primaryKey'][0] == 'concept':
                    ddf_schema['concepts'].append(sch_object)
                else:
                    ddf_schema['entities'].append(sch_object)
            else:
                if 'synonym' in sch['primaryKey']:
                    ddf_schema['synonyms'].append(sch_object)
                else:
                    ddf_schema['datapoints'].append(sch_object)

        return ddf_schema

    def get_ddf_schema(self, update=False):
        if not update and self.ddfSchema is not None:
            return self.ddfSchema
        elif not update and self.ddfSchema is None:
            raise ValueError('No ddfSchema, please use update=True to generate one')
        else:
            self.ddfSchema = self.generate_ddf_schema()
            return self.ddfSchema



# -*- coding: utf-8 -*-

from . extract_concepts import extract_concepts
from . filter import filter
from . flatten import flatten
from . groupby import groupby
from . merge import merge
from . merge_entity import merge_entity
from . run_op import run_op
from . split_entity import split_entity
from . translate_column import translate_column
from . translate_header import translate_header
from . trend_bridge import trend_bridge
from . window import window



# -*- coding: utf-8 -*-

"""functions working with patches"""

import pandas as pd
import daff


def apply_patch(base, patch):
    """apply patch created with daff.
    more on the diff format, see: http://specs.frictionlessdata.io/tabular-diff-format/

    return: the updated DataFrame.
    """
    d = daff.Coopy()

    base_df = pd.read_csv(base, header=None)
    patch_df = pd.read_csv(patch, header=None)

    base_list = base_df.values.tolist()
    if d.patch(base_list, patch_df.values.tolist()):
        new_df = pd.DataFrame(base_list[1:], columns=base_list[0])
    else:
        raise ValueError("patch failed. double check you patch")

    return new_df



# -*- coding: utf-8 -*-
"""io functions for ddf files"""

import os
import shutil
import json
import threading
import time
import typing
from urllib.parse import urlsplit

from io import BytesIO
import pandas as pd
import requests as req

from ddf_utils.str import format_float_digits
from ddf_utils.package import get_datapackage


# helper for dumping datapackage json
def dump_json(path, obj):
    """convenient function to dump a dictionary object to json"""
    with open(path, 'w+') as f:
        json.dump(obj, f, ensure_ascii=False, indent=4)
        f.close()


# TODO: integrate with Ingredient.serve
def serve_datapoint(df_: pd.DataFrame, out_dir, concept, copy=True,
                    by: typing.Iterable = None,
                    formatter: typing.Callable = format_float_digits, **kwargs):
    """save a pandas dataframe to datapoint file.
    the file path of csv will be out_dir/ddf--datapoints--$concept--$by.csv

    addition keyword arguments can be passed to `pd.DataFrame.to_csv()` function.
    """
    if copy:
        df = df_.copy()
    else:
        df = df_
    # formatting the concept column
    if formatter is not None:
        df[concept] = df[concept].map(formatter)

    if by is None:
        by = df.index.names
    by = '--'.join(by)

    path = os.path.join(out_dir, 'ddf--datapoints--{}--by--{}.csv'.format(concept, by))
    df.to_csv(path, **kwargs)


def serve_concept():
    pass


def serve_entity():
    pass


def open_google_spreadsheet(docid):
    """read google spreadsheet into excel io object"""
    tmpl_xls = "https://docs.google.com/spreadsheets/d/{docid}/export?format=xlsx&id={docid}"
    url = tmpl_xls.format(docid=docid)
    res = req.get(url)
    if res.ok:
        return BytesIO(res.content)
    return None


def cleanup(path, how='ddf'):
    """remove all ddf files in the given path"""
    # TODO: support names don't have standard ddf name format.
    if how == 'ddf':
        for f in os.listdir(path):
            if f.startswith("ddf--"):
                p = os.path.join(path, f)
                if os.path.isdir(p):
                    shutil.rmtree(p)
                else:
                    os.remove(p)
        # TODO: think a best way to handle metadata in datapackage.json
        # if os.path.exists(os.path.join(path, 'datapackage.json')):
        #     os.remove(os.path.join(path, 'datapackage.json'))
    if how == 'lang':
        if os.path.exists(os.path.join(path, 'lang')):
            shutil.rmtree(os.path.join(path, 'lang'))
    if how == 'langsplit':
        if os.path.exists(os.path.join(path, 'langsplit')):
            shutil.rmtree(os.path.join(path, 'langsplit'))


def download_csv(urls, out_path):
    """download csv files"""

    def download(url_, out_path_):
        r = req.get(url_, stream=True)
        total_length = int(r.headers.get('content-length'))
        if total_length == 0:
            return
        fn = urlsplit(url_).path.split('/')[-1]
        print('writing to: {}\n'.format(fn), end='')
        with open(os.path.join(out_path, fn), 'wb') as fd:
            for chunk in r.iter_content(chunk_size=128):
                fd.write(chunk)

    def create_tread(url_, out_path_):
        download_thread = threading.Thread(target=download, args=(url_, out_path_))
        download_thread.start()
        return download_thread

    threads = []
    for url in urls:
        threads.append(create_tread(url, out_path))

    # wait until all downloads are done
    is_alive = [t.is_alive() for t in threads]
    while any(is_alive):
        time.sleep(1)
        is_alive = [t.is_alive() for t in threads]


def csvs_to_ddf(files, out_path):
    """convert raw files to ddfcsv

    Args
    ----
    files: list
        a list of file paths to build ddf csv
    out_path: `str`
        the directory to put the ddf dataset

    """
    import re
    from os.path import join
    from ddf_utils.str import to_concept_id

    concepts_df = pd.DataFrame([['name', 'Name', 'string']],
                               columns=['concept', 'name', 'concept_type'])
    concepts_df = concepts_df.set_index('concept')

    all_entities = dict()

    pattern = r'indicators--by--([ 0-9a-zA-Z_-]*).csv'

    for f in files:
        data = pd.read_csv(f)
        basename = os.path.basename(f)
        keys = re.match(pattern, basename).groups()[0].split('--')
        keys_alphanum = list(map(to_concept_id, keys))

        # check if there is a time column. Assume last column is time.
        try:
            pd.to_datetime(data[keys[-1]], format='%Y')
        except (ValueError, pd.tslib.OutOfBoundsDatetime):
            has_time = False
        else:
            has_time = True

        if has_time:
            ent_keys = keys[:-1]
        else:
            ent_keys = keys

        # set concept type
        for col in data.columns:
            concept = to_concept_id(col)

            if col in keys:
                if col in ent_keys:
                    t = 'entity_domain'
                else:
                    t = 'time'
            else:
                t = 'measure'

            concepts_df.loc[concept] = [col, t]

        for ent in ent_keys:
            ent_df = data[[ent]].drop_duplicates().copy()
            ent_concept = to_concept_id(ent)
            ent_df.columns = ['name']
            ent_df[ent_concept] = ent_df.name.map(to_concept_id)

            if ent_concept not in all_entities.keys():
                all_entities[ent_concept] = ent_df
            else:
                all_entities[ent_concept] = pd.concat([all_entities[ent_concept], ent_df],
                                                      ignore_index=True)

        data = data.set_index(keys)
        for c in data:
            # output datapoints
            df = data[c].copy()
            df = df.reset_index()
            for k in keys[:-1]:
                df[k] = df[k].map(to_concept_id)
            df.columns = df.columns.map(to_concept_id)
            (df.dropna()
               .to_csv(join(out_path,
                            'ddf--datapoints--{}--by--{}.csv'.format(
                                to_concept_id(c), '--'.join(keys_alphanum))),
                       index=False))

    # output concepts
    concepts_df.to_csv(join(out_path, 'ddf--concepts.csv'))

    # output entities
    for c, df in all_entities.items():
        df.to_csv(join(out_path, 'ddf--entities--{}.csv'.format(c)), index=False)

    dp = get_datapackage(out_path, use_existing=False)
    dump_json(os.path.join(out_path, 'datapackage.json'), dp)

    return



# -*- coding: utf-8 -*-

"""main ingredient class"""

import numpy as np
from ..str import format_float_digits
from .helpers import read_opt
import os
import logging

from ..ddf_reader import DDF
from .exceptions import IngredientError


class BaseIngredient(object):
    def __init__(self, ingred_id, key, data=None):
        self.ingred_id = ingred_id
        self.key = key
        self.data = data

    @property
    def dtype(self):
        """
        returns the type of ddf data, i.e. concepts/entities/datapoints.

        It will be inferred from the key property of ingredient.
        """
        # TODO: what if key == '*'? Is it possible?
        keys = self.key_to_list()
        if len(keys) == 1:
            if keys[0] == 'concept':
                return 'concepts'
            else:
                return 'entities'
        else:
            return 'datapoints'

    def key_to_list(self):
        """helper function: make a list that contains primaryKey of this ingredient"""
        return [x.strip() for x in self.key.split(',')]

    def get_data(self):
        return self.data

    def copy_data(self):
        """this function makes copy of self.data.
        """
        if not self.data:
            self.get_data()
        # v: DataFrame. DataFrame.copy() is by default deep copy,
        # but I just call it explicitly here.
        return dict((k, v.copy(deep=True)) for k, v in self.data.items())

    def reset_data(self):
        self.data = None
        return

    def _serve_concepts(self, outpath, **options):
        data = self.copy_data()
        assert isinstance(data, dict)
        for k, df in data.items():
            # change boolean into string
            for i, v in df.dtypes.iteritems():
                if v == 'bool':
                    df[i] = df[i].map(lambda x: str(x).upper())
            path = os.path.join(outpath, 'ddf--concepts.csv')
            df.to_csv(path, index=False, encoding='utf8')

    def _serve_entities(self, outpath, **options):
        data = self.copy_data()
        assert isinstance(data, dict)
        for k, df in data.items():
            # change boolean into string
            for i, v in df.dtypes.iteritems():
                if v == 'bool':
                    df[i] = df[i].map(lambda x: str(x).upper())
            domain = self.key
            if k == domain:
                path = os.path.join(outpath, 'ddf--entities--{}.csv'.format(k))
            else:
                path = os.path.join(outpath, 'ddf--entities--{}--{}.csv'.format(domain, k))
            df.to_csv(path, index=False, encoding='utf8')

    def _serve_datapoints(self, outpath, **options):
        data = self.copy_data()
        assert isinstance(data, dict)
        digits = read_opt(options, 'digits', default=5)

        def to_disk(df_input, k, path):
            df = df_input.copy()
            if not np.issubdtype(df[k].dtype, np.number):
                try:
                    df[k] = df[k].astype(float)
                    df[k] = df[k].map(lambda x: format_float_digits(x, digits))
                except ValueError:
                    logging.warning("data not numeric: " + k)
            else:
                df[k] = df[k].map(lambda x: format_float_digits(x, digits))
            df.to_csv(path, encoding='utf8', index=False)

        for k, df in data.items():
            split_by = read_opt(options, 'split_datapoints_by', default=False)
            by = self.key_to_list()
            if not split_by:
                columns = [*by, k]
                path = os.path.join(
                    outpath,
                    'ddf--datapoints--{}--by--{}.csv'.format(k, '--'.join(by)))
                to_disk(df[columns], k, path)
            else:
                # split datapoints by entities. Firstly we calculate all possible
                # combinations of entities, and then filter the dataframe, create
                # file names and save them into disk.
                from itertools import product
                values = list()
                [values.append(df[col].unique()) for col in split_by]
                all_combinations = product(*values)

                if len(by) > len(split_by):
                    by = list(set(by) - set(split_by))
                    columns = [*sorted(split_by), *sorted(by), k]
                else:
                    by = None
                    columns = [*sorted(split_by), k]

                for comb in all_combinations:
                    query = ''
                    entity_strings = list()
                    for entity, value in zip(split_by, comb):
                        entity_strings.append(entity + '-' + value)
                        if len(query) > 0:
                            query = query + " and " + "{} == '{}'".format(entity, value)
                        else:
                            query = "{} == '{}'".format(entity, value)

                    if by:
                        path = os.path.join(
                            outpath,
                            'ddf--datapoints--{}--by--{}--{}.csv'.format(
                                k,
                                '--'.join(sorted(entity_strings)),
                                '--'.join(sorted(by))
                            )
                        )
                    else:
                        path = os.path.join(
                            outpath,
                            'ddf--datapoints--{}--by--{}.csv'.format(
                                k,
                                '--'.join(sorted(entity_strings))
                            )
                        )
                    # logging.debug('query is: ' + query)
                    df_part = df.query(query)
                    to_disk(df_part[columns], k, path)

    def serve(self, outpath, **options):
        """save the ingledient to disk.

        Parameters
        ----------
        outpath : str
            the path to save the ingredient

        Other Parameters
        ----------------
        digits : int
            how many digits to keep at most.

        """
        logging.info('serving ingredient: ' + self.ingred_id)
        # create outpath if not exists
        if 'sub_folder' in options:
            sub_folder = options.pop('sub_folder')
            assert not os.path.isabs(sub_folder)  # sub folder should not be abspath
            outpath = os.path.join(outpath, sub_folder)
        os.makedirs(outpath, exist_ok=True)

        t = self.dtype
        if t == 'datapoints':
            self._serve_datapoints(outpath, **options)
        elif t == 'concepts':
            self._serve_concepts(outpath, **options)
        elif t == 'entities':
            self._serve_entities(outpath, **options)
        else:
            raise IngredientError('Not a correct collection: ' + t)


class Ingredient(BaseIngredient):
    """
    ingredient class: represents an ingredient object in recipe file.
    see the implement of from_dict() method for how the object is constructed.

    Here is an example ingredient object in recipe:

    .. code-block:: yaml

        id: example-ingredient
        dataset: ddf--example--dataset
        key: "geo,time"  # key columns of ingredient
        value:  # only include concepts listed here
          - concept_1
          - concept_2
        filter:  # select rows by column values
          geo:  # only keep datapoint where `geo` is in [swe, usa, chn]
            - swe
            - usa
            - chn

    Attributes
    ----------
    ingred_id : `str`
        The id string
    ddf_id : `str`
        the underlaying dataset id
    key : `str`
        the key columns of the ingredient
    value : `list`
        concept filter applied to the dataset. if `value` == "*", all concept of
        the dataset will be in the ingredient
    row_filter : `dict`
        row filter applied to the dataset

    Methods
    -------
    get_data()
        read in and return the ingredient data

    """
    def __init__(self, ingred_id,
                 ddf_id=None, key=None, values=None, row_filter=None, data=None):
        super(Ingredient, self).__init__(ingred_id, key, data)
        self.values = values
        self.row_filter = row_filter
        self._ddf_id = ddf_id
        self._ddf = None

    @classmethod
    def from_dict(cls, data):
        """create an instance by a dictionary

        The dictionary should provide following keys:

        - id
        - dataset
        - key
        - value
        - filter (optional)
        """
        ingred_id = data['id']
        ddf_id = data['dataset']
        key = data['key']
        values = data['value']
        if 'row_filter' in data.keys():
            row_filter = data['row_filter']
        else:
            row_filter = None

        return cls(ingred_id, ddf_id, key, values, row_filter)

    @property
    def ddf(self):
        """The DDF reader object"""
        if self._ddf:
            return self._ddf
        else:
            if self._ddf_id:
                self._ddf = DDF(self._ddf_id)
                return self._ddf
        return None

    @property
    def ddf_path(self):
        """The path to the dataset"""
        return self.ddf.dataset_path

    def __repr__(self):
        return '<Ingredient: {}>'.format(self.ingred_id)

    def _get_data_datapoint(self, copy):
        data = dict()
        keys = self.key_to_list()
        if self.values == '*':  # get all datapoints for the key
            values = []
            for k, v in self.ddf.get_datapoint_files().items():
                for pkeys in v.keys():
                    if set(keys) == set(pkeys):
                        values.append(k)
        else:
            values = self.values

        if not values or len(values) == 0:
            raise IngredientError('no datapoint found for the ingredient: ' + self.ingred_id)

        for v in values:
            data[v] = self.ddf.get_datapoint_df(v, keys)
        return data

    def _get_data_entities(self, copy):
        # TODO: values should always be '*'?
        if copy:
            ent = self.ddf.get_entities(dtype=str)
        else:
            ent = self.ddf.get_entities()
        conc = self.ddf.get_concepts()
        values = []

        if conc.ix[self.key, 'concept_type'] == 'entity_domain':
            if 'domain' in conc.columns and len(conc[conc['domain'] == self.key]) > 0:
                [values.append(i) for i in conc[conc['domain'] == self.key].index]
            else:
                values.append(self.key)
        else:
            values.append(self.key)

        return dict((k, ent[k]) for k in values)

    def _get_data_concepts(self, copy):
        if self.values == '*':
            return {'concepts': self.ddf.get_concepts()}
        else:
            return {'concepts': self.ddf.get_concepts()[self.values]}

    def get_data(self, copy=False, key_as_index=False):
        """read in and return the ingredient data
        """
        funcs = {
            'datapoints': self._get_data_datapoint,
            'entities': self._get_data_entities,
            'concepts': self._get_data_concepts
        }

        def filter_row(df):
            """return the rows selected by self.row_filter."""
            # TODO: improve filtering function
            # 1. know more about the row_filter syntax
            # 2. The query() Method is Experimental
            if self.row_filter:
                query_str = "and".join(["{} in {}".format(k, v) for k, v in self.row_filter.items()])
                # logging.debug(query_str)
                df = df.query(query_str)
                # logging.debug(df.head())
            return df

        if self.data is None:
            data = funcs[self.dtype](copy)
            for k, v in data.items():
                if self.row_filter:
                    # index_cols = data[k].index.names
                    # data[k] = self.filter_row(data[k].reset_index()).set_index(index_cols)
                    data[k] = filter_row(data[k].reset_index())
                else:
                    data[k] = data[k].reset_index()
            self.data = data
        if key_as_index:
            # TODO set index when requiring data
            pass
        return self.data


class ProcedureResult(BaseIngredient):
    def __init__(self, ingred_id, key, data):
        super(ProcedureResult, self).__init__(ingred_id, key, data)

    def __repr__(self):
        return '<ProcedureResult: {}>'.format(self.ingred_id)

    def reset_data(self):
        # TODO: allowing reset data? It can not be reconstructed.
        raise NotImplementedError('')



# -*- coding: utf-8 -*-

__version__ = '1.0.1'

from . import (str, cli, i18n, io, patch, qa, transformer, factory)
from ddf_utils.model.package import DataPackage
from ddf_utils.package import get_datapackage, create_datapackage



# -*- coding: utf-8 -*-

"""filter procedure for recipes"""


import logging

from typing import List, Sequence
from .. exceptions import ProcedureError
from .. helpers import debuggable, read_opt, query
from .. model.ingredient import Ingredient, get_ingredient_class
from .. model.chef import Chef


logger = logging.getLogger('filter')


def _check_item_type(item):
    """item filter only accept certain format"""
    if isinstance(item, Sequence):
        return "list"
    elif isinstance(item, dict) and len(item) == 1:
        assert list(item.keys())[0] in ['$in', '$nin'], "only $in, $nin supported in item filter"
        return "mongo"
    else:
        raise ProcedureError("not supported item filter: " + str(item))


def _filter_datapoint_item(ingredient, item):

    data = ingredient.get_data()

    def filter_in_list(d, l):
        r = dict()
        for i in l:
            if i in d.keys():
                r[i] = d[i].copy()
            else:
                logger.warning("concept {} not found in ingredient {}".format(i, ingredient.id))
        return r

    item_type = _check_item_type(item)
    if item_type == 'list':
        res = filter_in_list(data, item)
    else:  # mongo
        selector = list(item.keys())[0]
        item_list = list(item.values())[0]
        if selector == '$in':
            res = filter_in_list(data, item_list)
        else:  # $nin
            res = dict()
            for k, df in data.items():
                if k not in item_list:
                    res[k] = data[k].copy()
    return res


def _filter_other_item(ingredient, item):
    data = ingredient.get_data()
    item_type = _check_item_type(item)

    def filter_in_list(d, l):
        r = dict()
        for k, v in d.items():
            items_ = l.copy()
            for i in items_:
                if i not in v.columns:
                    items_.remove(i)
                    logger.warning("concept {} not found in ingredient {}".format(i, ingredient.id))
            r[k] = v[items_].copy()
        return r

    if item_type == 'list':
        res = filter_in_list(data, item)
    else:  # mongo
        selector = list(item.keys())[0]
        item_list = list(item.values())[0]
        if selector == '$in':
            res = filter_in_list(data, item_list)
        else:
            res = dict()
            for k, v in data.items():
                for i in item_list:
                    if i not in v.columns:
                        logger.warning("concept {} not found in ingredient {}".format(i, ingredient.id))
                keep_cols = list(set(v.columns.values) - set(item_list))
                res[k] = v[keep_cols].copy()
    return res


@debuggable
def filter(chef: Chef, ingredients: List[Ingredient], result, **options) -> Ingredient:
    """filter items and rows just as what `value` and `filter` do in ingredient definition.

    Procedure format:

    .. code-block:: yaml

       - procedure: filter
         ingredients:
             - ingredient_id
         options:
             item:  # just as `value` in ingredient def
                 $in:
                     - concept_1
                     - concept_2
             row:  # just as `filter` in ingredient def
                 $and:
                     geo:
                         $ne: usa
                     year:
                         $gt: 2010

         result: output_ingredient

    for more information, see the :py:class:`ddf_utils.chef.ingredient.Ingredient` class.

    Parameters
    ----------
    chef: Chef
        the Chef instance
    ingredients:
        list of ingredient id in the DAG
    result: `str`

    Keyword Args
    ------------
    item: list or dict, optional
        The item filter
    row: dict, optional
        The row filter
    """

    assert len(ingredients) == 1, "procedure only support 1 ingredient for now."
    # ingredient = chef.dag.get_node(ingredients[0]).evaluate()
    ingredient = ingredients[0]
    logger.info("filter_row: " + ingredient.id)

    row_filters = read_opt(options, 'row', False, None)
    items = read_opt(options, 'item', False, None)

    if row_filters is None and items is None:
        raise ProcedureError('filter procedure: at least one of `row` and `item` should be set in the options!')

    # first handel item filter, then row filter.
    # easily to see that the order doesn't affect the final result.
    if items is not None and len(items) != 0:
        if ingredient.dtype == 'datapoints':
            res = _filter_datapoint_item(ingredient, items)
        else:
            res = _filter_other_item(ingredient, items)
    else:
        res = dict()
        data = ingredient.get_data()
        for k, df in data.items():
            res[k] = df.copy()

    if row_filters is not None:
        for k, df in res.items():
            res[k] = query(df, row_filters, available_scopes=df.columns)

    return get_ingredient_class(ingredient.dtype).from_procedure_result(result, ingredient.key, data_computed=res)



# -*- coding: utf-8 -*-
"""recipe cooking"""

import json
import yaml
import re
import os
from orderedattrdict import AttrDict
from orderedattrdict.yamlutils import AttrDictYAMLLoader

from . ingredient import Ingredient
from . dag import DAG, IngredientNode, ProcedureNode
from . helpers import read_opt
from .. import config
from . exceptions import ChefRuntimeError

import logging

logger = logging.getLogger('Chef')


def _loadfile(f):
    """load json/yaml file, into AttrDict"""
    if re.match('.*\.json', f):
        res = json.load(open(f), object_pairs_hook=AttrDict)
    else:
        res = yaml.load(open(f), Loader=AttrDictYAMLLoader)

    return res


# functions for reading/running recipe
def build_recipe(recipe_file, to_disk=False, **kwargs):
    """build a complete recipe object.

    This function will check each part of recipe, convert string (the ingredient ids,
    dictionaries file names) into actual objects.

    If there are includes in recipe file, this function will run recurivly.
    If no includes found then return the parsed object as is.

    Parameters
    ----------
    recipe_file : `str`
        path to recipe file

    Keyword Args
    ------------
    to_disk : bool
        if true, save the parsed reslut to a yaml file in working dir

    Other Parameters
    ----------------
    ddf_dir : `str`
        path to search for DDF datasets, will overwrite the contfig in recipe

    """
    recipe = _loadfile(recipe_file)

    # the base dir of recipe file. for building paths for dictionary_dir and
    # sub recipe paths.
    base_dir = os.path.dirname(recipe_file)

    # the dictionary dir to retrieve translation dictionaries
    try:
        dict_dir = recipe['config']['dictionary_dir']
        config.DICT_PATH = dict_dir
    except KeyError:
        dict_dir = config.DICT_PATH

    # expand all files in the options
    if 'cooking' in recipe.keys():
        for p in ['concepts', 'datapoints', 'entities']:
            if p not in recipe['cooking'].keys():
                continue
            for i, procedure in enumerate(recipe['cooking'][p]):
                try:
                    opt_dict = procedure['options']['dictionary']
                except KeyError:
                    continue
                if isinstance(opt_dict, str):
                    # if the option dict is str, then it should be a filename
                    if dict_dir is None:
                        raise ChefRuntimeError("dictionary_dir not found in config!")
                    if os.path.isabs(dict_dir):
                        path = os.path.join(dict_dir, opt_dict)
                    else:
                        path = os.path.join(base_dir, dict_dir, opt_dict)

                    recipe['cooking'][p][i]['options']['dictionary'] = _loadfile(path)

    # setting ddf search path if option is provided
    if 'ddf_dir' in kwargs.keys():
        if 'config' not in recipe.keys():
            recipe['config'] = AttrDict()
        recipe.config.ddf_dir = kwargs['ddf_dir']
        config.DDF_SEARCH_PATH = kwargs['ddf_dir']

    if 'include' not in recipe.keys():
        return recipe
    else:  # append sub-recipe entities into main recipe
        recipe_dir = recipe['config']['recipes_dir']

        sub_recipes = []
        for i in recipe['include']:
            # TODO: maybe add support to expand user home and env vars
            if os.path.isabs(recipe_dir):
                path = os.path.join(recipe_dir, i)
            else:
                path = os.path.join(base_dir, recipe_dir, i)
            sub_recipes.append(build_recipe(path))

        for rcp in sub_recipes:
            # appending ingredients
            if 'ingredients' in recipe.keys():
                # ingredients = [*recipe['ingredients'], *rcp['ingredients']]
                # ^ not supportted by Python < 3.5
                ingredients = []
                if 'ingredients' in recipe.keys():
                    [ingredients.append(ing) for ing in recipe['ingredients']]
                if 'ingredients' in rcp.keys():
                    [ingredients.append(ing) for ing in rcp['ingredients']]
                # drop duplicated ingredients.
                rcp_dict_tmp = {}
                for v in ingredients:
                    if v['id'] not in rcp_dict_tmp.keys():
                        rcp_dict_tmp[v['id']] = v
                    else:
                        # raise error when ingredients with same ID have different contents.
                        if v != rcp_dict_tmp[v['id']]:
                            raise ChefRuntimeError(
                                "Different content with same ingredient id detected: " + v['id'])
                recipe['ingredients'] = list(rcp_dict_tmp.values())
            else:
                recipe['ingredients'] = rcp['ingredients']

            # appending cooking procedures
            if 'cooking' not in rcp.keys():
                continue
            for p in ['datapoints', 'entities', 'concepts']:
                if p not in rcp['cooking'].keys():
                    continue
                if 'cooking' in recipe.keys():
                    if p in recipe['cooking'].keys():
                        # NOTE: the included cooking procedures should be placed in front of
                        # the origin ones.
                        # recipe['cooking'][p] = [*rcp['cooking'][p], *recipe['cooking'][p]]
                        # ^ not supportted by Python < 3.5
                        new_procs = []
                        [new_procs.append(proc) for proc in rcp['cooking'][p]]
                        [new_procs.append(proc) for proc in recipe['cooking'][p]
                         if proc not in new_procs]
                        recipe['cooking'][p] = new_procs
                    else:
                        recipe['cooking'][p] = rcp['cooking'][p]
                else:
                    recipe['cooking'] = {}
                    recipe['cooking'][p] = rcp['cooking'][p]

        return recipe


def update_recipe_last_update(recipe, outdir):
    pass


def check_dataset_availability(recipe):
    """check availability of all datasets required by the recipe.

    raise error if some dataset not available. otherwise do nothing.
    """
    try:
        ddf_dir = recipe.config.ddf_dir
    except (KeyError, AttributeError):
        ddf_dir = config.DDF_SEARCH_PATH

    datasets = set()
    for ingred in recipe.ingredients:
        datasets.add(ingred.dataset)

    not_exists = []
    for d in datasets:
        if not os.path.exists(os.path.join(ddf_dir, d)):
            not_exists.append(d)

    if len(not_exists) > 0:
        logging.critical("not enough datasets! please checkout following datasets:\n{}\n"
                         .format('\n'.join(not_exists)))
        raise ChefRuntimeError('not enough datasets')
    return


def build_dag(recipe):
    """build a DAG model for the recipe.

    For more detail for DAG model, see :py:mod:`ddf_utils.chef.dag`.
    """

    def add_dependency(dag, upstream_id, downstream):
        if not dag.has_node(upstream_id):
            upstream = ProcedureNode(upstream_id, None, dag)
            dag.add_node(upstream)
        else:
            upstream = dag.get_node(ing)
        dag.add_dependency(upstream.node_id, downstream.node_id)

    dag = DAG()
    ingredients = [Ingredient.from_dict(x) for x in recipe.ingredients]
    # ingredients_names = [x.ingred_id for x in ingredients]
    serving = []

    # adding ingredient nodes
    for i in ingredients:
        dag.add_node(IngredientNode(i.ingred_id, i, dag))

    for k, d in recipe.cooking.items():
        for proc in d:
            if proc['procedure'] == 'serve':
                [serving.append(x) for x in proc.ingredients]
                continue
            try:
                node = ProcedureNode(proc.result, proc, dag)
            except KeyError:
                logger.critical('Please set the result id for procedure: ' + proc['procedure'])
                raise
            dag.add_node(node)
            # add nodes from base ingredients
            for ing in proc.ingredients:
                add_dependency(dag, ing, node)
            # also add nodes from options
            # for now, if a key in option named base or ingredient, it will be treat as ingredients
            # TODO: see if there is better way to do
            if 'options' in proc.keys():
                options = proc['options']
                for ingredient_key in ['base', 'ingredient']:
                    if ingredient_key in options.keys():
                        add_dependency(dag, options[ingredient_key], node)
                    for opt, val in options.items():
                        if isinstance(val, AttrDict) and ingredient_key in val.keys():
                            add_dependency(dag, options[opt][ingredient_key], node)
            # detect cycles in recipe after adding all related nodes
            node.detect_downstream_cycle()
    # check if all serving ingredients are available
    for i in serving:
        if not dag.has_node(i):
            raise ChefRuntimeError('Ingredient not found: ' + i)
    if 'serving' in recipe.keys():
        if len(serving) > 0:
            raise ChefRuntimeError('can not have serve procedure and serving section at same time!')
        for i in recipe['serving']:
            if not dag.has_node(i['id']):
                raise ChefRuntimeError('Ingredient not found: ' + i['id'])
    # display the tree
    # dag.tree_view()
    return dag


def get_dishes(recipe):
    """get all dishes in the recipe"""

    if 'serving' in recipe:
        return recipe['serving']

    dishes = list()
    for _, procs in recipe['cooking'].items():
        serve_proc_exists = False
        for p in procs:
            if p['procedure'] == 'serve':
                serve_proc_exists = True
                for i in p['ingredients']:
                    try:
                        dishes.append({'id': i, 'options': p['options']})
                    except KeyError:
                        dishes.append({'id': i, 'options': dict()})
        if not serve_proc_exists:
            logger.warning('no serve procedure found, serving the last result: ' + p['result'])
            dishes.append({'id': p['result'], 'options': dict()})

    return dishes


def run_recipe(recipe, serve=False, outpath=None):
    """run the recipe.

    returns a dictionary. keys are `concepts`, `entities` and `datapoints`,
    and values are ingredients defined in the `serve` procedures or `serving` section.
    for example:

    .. code-block:: python

        {
            "concepts": [{"ingredient": DataFrame1, "options": None}]
            "datapoints": [
                {
                    "ingredient": DataFrame2,
                    "options": {"digits": 5}
                },
                {
                    "ingredient": DataFrame3,
                    "options": {"digits": 1}
                },
            ]
        }

    """
    try:
        config.DDF_SEARCH_PATH = recipe['config']['ddf_dir']
    except KeyError:
        if not config.DDF_SEARCH_PATH:
            raise ChefRuntimeError("no ddf_dir configured, please check your recipe")
    logging.info('path for searching DDF: ' + str(config.DDF_SEARCH_PATH))

    # make a copy recipe, don't change the origin recipe
    from copy import deepcopy
    recipe_ = deepcopy(recipe)

    # check all datasets availability
    check_dataset_availability(recipe_)

    # create DAG of recipe
    dag = build_dag(recipe_)

    # check all ingredients availability
    for root in dag.roots:
        root.detect_missing_dependency()

    # now run the recipe
    dishes = get_dishes(recipe_)

    results = list()
    for dish in dishes:
        dish_result = dag.get_node(dish['id']).evaluate()
        results.append(dish_result)
        if serve:
            if 'options' in dish:
                dish_result.serve(outpath, **dish['options'])
            else:
                dish_result.serve(outpath)

    return results



# -*- coding: utf-8 -*-

"""split_entity procedure for recipes"""

import logging
from typing import List

from .. helpers import debuggable
from .. model.ingredient import DataPointIngredient
from .. model.chef import Chef


logger = logging.getLogger('split_entity')


@debuggable
def split_entity(chef: Chef, ingredients: List[DataPointIngredient], dictionary,
                 target_column, result, splitted='drop') -> DataPointIngredient:
    """split entities"""
    from ... transformer import split_keys

    assert len(ingredients) == 1, "procedure only support 1 ingredient for now."
    # ingredient = chef.dag.get_node(ingredients[0]).evaluate()
    ingredient = ingredients[0]

    data = ingredient.compute()

    res_data = dict()
    for k, df in data.items():
        res_data[k] = split_keys(df.set_index(ingredient.key),
                                 target_column, dictionary, splitted).reset_index()

    return DataPointIngredient.from_procedure_result(result, ingredient.key, res_data)



# -*- coding: utf-8 -*-

"""
model for dataset repositories.
"""

import logging
import os.path as osp
import tempfile
from urllib.parse import urlparse

from git import Repo as GRepo

from .package import DataPackage

logger = logging.getLogger(__name__)


def is_url(r):
    if r.startswith('https://') or r.startswith('git://'):
        return True
    else:
        return False


class Repo:
    def __init__(self, uri, base_path=None):
        if is_url(uri):
            if base_path is None:
                base_path = tempfile.mkdtemp()
            sub_path = osp.splitext(urlparse(uri).path)[0][1:]

            # self._name = osp.basename(sub_path)
            self._name = sub_path
            self._local_path = osp.join(base_path, sub_path)
            if osp.exists(self._local_path):
                # TODO: maybe give error?
                self.__repo = GRepo(self._local_path)
            else:
                print('checking out repo {} to {}'.format(sub_path, base_path))
                self.__repo = GRepo.clone_from(uri, to_path=self._local_path)

        else:
            assert osp.exists(osp.abspath(uri)), 'the path "{}" does not exist'.format(uri)
            self._name = osp.basename(uri)
            self._local_path = uri
            self.__repo = GRepo(uri)

    @property
    def name(self):
        return self._name

    @property
    def local_path(self):
        return self._local_path

    def show_versions(self):
        if len(self.__repo.tags) > 0:
            print('available versions:')
            [print(tag.name) for tag in self.__repo.tags]
        else:
            print("no version for the repo yet")

    def to_datapackage(self, ref=None):
        """turn repo@ref into Datapackage"""
        if ref is not None:
            self.__repo.git.checkout(ref)

        return DataPackage(self.local_path)



# -*- coding: utf-8 -*-
#!/usr/bin/env python

"""script for ddf dataset management tasks"""

import logging
import os
import shutil

import click
import coloredlogs

LOG_LEVEL = logging.INFO


@click.group()
@click.option('--debug/--no-debug', default=False)
def ddf(debug):
    global LOG_LEVEL
    if debug:
        LOG_LEVEL = logging.DEBUG
    # logging.basicConfig(level=level, format='%(asctime)s -%(levelname)s- %(message)s',
    #                     datefmt="%H:%M:%S"
    #                     )
    coloredlogs.install(level=LOG_LEVEL, fmt='%(asctime)s %(levelname)s %(message)s')


# project management
@ddf.command()
def new():
    """create a new ddf project"""
    from cookiecutter.main import cookiecutter
    cookiecutter('https://github.com/semio/ddf_project_template')


@ddf.command()
@click.argument('how', default='ddf', type=click.Choice(['ddf', 'lang', 'langsplit']))
@click.argument('path', default='./')
@click.option('--force', flag_value=True, default=False, help='force deletion')
def cleanup(path, how, force):
    """clean up ddf files or translation files.

    :arguments

    \b
    how: what to clean, choose from 'ddf', 'lang' and 'langsplit'
    path: the dataset path
    """
    from ddf_utils.io import cleanup as cl
    from ddf_utils.package import is_datapackage
    if force:
        cl(path, how)
    else:
        if not is_datapackage(path):
            print('not a dataset path: {}. Please set correct path or '
                  'use --force to force run.'.format(os.path.abspath(path)))
        else:
            cl(path, how)
    click.echo('Done.')


@ddf.command()
@click.argument('path')
@click.option('--update', '-u', 'update', flag_value=True, default=False,
              help='update existing datapackage.json')
@click.option('--overwrite', '-n', 'overwrite', flag_value=True, default=False,
              help='overwrite existing datapackage.json')
def create_datapackage(path, update, overwrite):
    """create datapackage.json"""
    from ddf_utils.package import get_datapackage
    import json
    if not update and not overwrite:
        if os.path.exists(os.path.join(path, 'datapackage.json')):
            click.echo('datapackage.json already exists. use --update to update or --overwrite to create new')
            return
        res = get_datapackage(path, use_existing=False)
    else:
        if os.path.exists(os.path.join(path, 'datapackage.json')):
            click.echo('backing up previous datapackage.json...')
            # make a backup
            shutil.copy(os.path.join(path, 'datapackage.json'),
                        os.path.join(path, 'datapackage.json.bak'))
        if overwrite:
            res = get_datapackage(path, use_existing=False)
        else:
            res = get_datapackage(path, use_existing=True, update=True)

    with open(os.path.join(path, 'datapackage.json'), 'w', encoding='utf8') as f:
        json.dump(res, f, indent=4, ensure_ascii=False)
    click.echo('Done.')


# chef and recipe
@ddf.command()
@click.option('--recipe', '-i', type=click.Path(exists=True), required=True)
@click.option('--outdir', '-o', type=click.Path(exists=True))
@click.option('--ddf_dir', type=click.Path(exists=True), default=None)
@click.option('--update', 'update', flag_value=False, help="Don't use. Not implemented yet")
@click.option('--dry_run', '-d', 'dry_run', flag_value=True, default=False,
              help="don't save output to disk")
@click.option('--no-gen-dp', '-p', 'gen_dp', flag_value=False, default=True,
              help="generate datapackage.json after recipe run")
@click.option('--show-tree', 'show_tree', flag_value=True, default=False,
              help='show the dependency tree')
def run_recipe(recipe, outdir, ddf_dir, update, dry_run, gen_dp, show_tree):
    """generate new ddf dataset with recipe"""
    from ddf_utils.chef.api import Chef
    from ddf_utils.package import create_datapackage
    from ddf_utils.io import dump_json
    import json

    coloredlogs.install(logger=logging.getLogger('Chef'),
                        fmt='%(asctime)s %(name)s %(levelname)s %(message)s',
                        level=LOG_LEVEL)

    click.echo('building recipe...')
    if ddf_dir:
        chef = Chef.from_recipe(recipe, ddf_dir=ddf_dir)
    else:
        chef = Chef.from_recipe(recipe)
    if show_tree:
        chef.dag.tree_view()
        return
    if update:
        pass
    serve = not dry_run
    chef.run(serve=serve, outpath=outdir)
    if serve and gen_dp:
        click.echo('creating datapackage file...')
        datapackage_path = os.path.join(outdir, 'datapackage.json')
        if os.path.exists(datapackage_path):
            click.echo('backup old datapackage.json to datapackage.json.bak')
            shutil.copyfile(datapackage_path, os.path.join(outdir, 'datapackage.json.bak'))
            dp_old = json.load(open(datapackage_path))
            # copy translations info. other info should be in the recipe.
            if 'translations' in dp_old.keys():
                chef = chef.add_metadata(translations=dp_old['translations'])
        dump_json(os.path.join(outdir, 'datapackage.json'),
                  create_datapackage(outdir, gen_schema=True, **chef.metadata))
    click.echo("Done.")


@ddf.command()
@click.argument('recipe', type=click.Path(exists=True))
@click.option('--format', '-f', type=click.Choice(['json', 'yaml']), default='json',
              help='set output format')
def build_recipe(recipe, format):
    """create a complete recipe by expanding all includes in the input recipe."""
    from ddf_utils.chef.api import Chef
    chef = Chef.from_recipe(recipe)
    fp = click.open_file('-', 'w')
    if format == 'json':
        import json
        json.dump(recipe, fp, indent=4, ensure_ascii=False)
    elif format == 'yaml':
        import yaml
        yaml.dump(recipe, fp)


# Translation related tasks
@ddf.command()
@click.argument('path', type=click.Path(exists=True))
@click.option('--type', '-t', 'dtype', type=click.Choice(['csv', 'json']), help='split file type',
              default='csv')
@click.option('--overwrite/--no-overwrite', default=False, help='overwrite existing files or not')
@click.option('--split_path', default='langsplit', help='path to langsplit folder')
@click.option('--exclude_concept', '-x', 'exclude_concepts', multiple=True,
              help='concepts to exclude', metavar='concept')
def split_translation(path, split_path, dtype, exclude_concepts, overwrite):
    """split ddf files for crowdin translation"""
    from ddf_utils.i18n import split_translations_csv, split_translations_json
    if dtype == 'csv':
        split_translations_csv(path, split_path, exclude_concepts, overwrite)
    elif dtype == 'json':
        split_translations_json(path, split_path, exclude_concepts, overwrite)
    else:
        click.echo('Please specify correct input type (csv or json) with -t option.')
        return
    click.echo('Done.')


@ddf.command()
@click.argument('path', type=click.Path(exists=True))
@click.option('--overwrite/--no-overwrite', default=False, help='overwrite existing files or not')
@click.option('--type', '-t', 'dtype', type=click.Choice(['json', 'csv']), help='split file type')
@click.option('--split_path', default='langsplit', help='path to langsplit folder')
@click.option('--lang_path', default='lang', help='path to lang folder')
def merge_translation(path, split_path, lang_path, dtype, overwrite):
    """merge all translation files from crowdin"""
    from ddf_utils.i18n import merge_translations_csv, merge_translations_json
    if dtype == 'csv':
        merge_translations_csv(path, split_path, lang_path, overwrite)
    elif dtype == 'json':
        merge_translations_json(path, split_path, lang_path, overwrite)
    else:
        click.echo('Please specify correct input type (csv or json) with -t option.')
        return
    click.echo('Done.')


# for QA
@ddf.command()
@click.argument('dataset1')
@click.argument('dataset2')
@click.option('--git', '-g', is_flag=True)
@click.option('--checkout-path', '-o', type=click.Path(), default='./etl/diff')
#  @click.option('--diff-only', is_flag=True)  # buggy
@click.option('--indicator', '-i', multiple=True)
@click.option('--on_key', '-k')  # for those comparsion need to group by keys
@click.option('-C', 'git_base_path', type=click.Path(), default='./')
def diff(dataset1, dataset2, git, checkout_path, indicator, on_key, git_base_path):
    """give a report on the statistical differences for datapoints between 2 datasets."""
    from ddf_utils.package import DDFcsv, is_datapackage
    from ddf_utils.qa import compare_with_func

    from os.path import join

    if git:
        from subprocess import check_output
        assert is_datapackage(git_base_path), 'please run this command in a dataset dir.'

        c1 = check_output(['git', '-C', git_base_path, 'rev-parse', dataset1])
        p1 = c1.strip().decode('utf8')

        c2 = check_output(['git', '-C', git_base_path, 'rev-parse', dataset2])
        p2 = c2.strip().decode('utf8')

        try:
            os.makedirs(join(checkout_path, p1))
            logging.info('checkout git rev {} into {}'.format(dataset1, join(checkout_path, p1)))
            os.system('git -C {} checkout {}'.format(git_base_path, p1))
            os.system('git -C {} checkout-index -a -f --prefix={}/'.format(git_base_path, join(checkout_path, p1)))
        except FileExistsError:
            pass

        try:
            os.makedirs(join(checkout_path, p2))
            logging.info('checkout git rev {} into {}'.format(dataset2, join(checkout_path, p2)))
            os.system('git -C {} checkout {}'.format(git_base_path, p2))
            os.system('git -C {} checkout-index -a -f --prefix={}/'.format(git_base_path, join(checkout_path, p2)))
        except FileExistsError:
            pass

        d1 = DDFcsv.from_path(join(checkout_path, p1)).ddf
        d2 = DDFcsv.from_path(join(checkout_path, p2)).ddf

    else:
        d1 = DDFcsv.from_path(dataset1).ddf
        d2 = DDFcsv.from_path(dataset2).ddf

    result = compare_with_func(d1, d2, fns=indicator, on=on_key)

    # sort it
    result = result.sort_values(by='indicator', ascending=True).set_index('indicator')

    # TODO: add an output type parameter, to choose between csv and tabulate.
    # click.echo(tabulate.tabulate(result,
    #                              headers=cols, tablefmt='psql'))
    click.echo(result.to_csv())


# csv to ddfcsv
@ddf.command()
@click.option('-i', 'input', type=click.Path(exists=True), default='./')
@click.option('-o', 'out_path', type=click.Path(exists=True))
def from_csv(input, out_path):
    """create ddfcsv dataset from a set of csv files"""
    from .io import csvs_to_ddf

    if os.path.isfile(input):
        files = [input]
    else:
        files = [os.path.join(input, x)
                 for x in os.listdir(input) if x.endswith('.csv')]

    csvs_to_ddf(files, out_path)

    click.echo('Done.')


# detect etl script type for a dataset
@ddf.command()
@click.option('-d', 'script_dir', type=click.Path(), default='etl/scripts')
def etl_type(script_dir):
    import sys

    script_dir = os.path.expanduser(script_dir)

    if not os.path.isabs(script_dir):
        script_dir = os.path.join('./', script_dir)

    if not os.path.exists(script_dir):
        raise FileNotFoundError('{} not exists!'.format(script_dir))

    sys.path.insert(0, script_dir)
    try:
        import etl
        fn = etl.recipe_file
        tp = 'recipe'
    except AttributeError:
        fn = ''
        tp = 'python'
    except ModuleNotFoundError:
        fn = ''
        tp = 'manual'
    print ('{},{}'.format(tp, fn))
    return


if __name__ == '__main__':
    ddf()



# -*- coding: utf-8 -*-

"""functions for scraping data from clio infra website.

Source link: `Clio-infra website`_

.. _`Clio-infra website`: https://www.clio-infra.eu/index.html
"""

import os.path as osp
import pandas as pd
from lxml import etree
import requests
from urllib.parse import urljoin

from .common import DataFactory


class ClioInfraLoader(DataFactory):

    url = 'https://clio-infra.eu/index.html'

    def _get_home_page(self, url):
        response = requests.get(url, verify=False)
        content = response.content
        tree = etree.fromstring(content, parser=etree.HTMLParser())

        return tree

    def has_newer_source(self, ver):
        print('there is no version info in this site.')
        raise NotImplementedError

    def load_metadata(self):
        tree = self._get_home_page(self.url)
        elem = tree.xpath('//div[@class="col-sm-4"]/div[@class="list-group"]/p[@class="list-group-item"]')

        res1 = {}
        res2 = {}

        for e in elem:
            try:
                name = e.find('a').text
                link = e.find('*/a').attrib['href']
                if '../data' in link:  # it's indicator file
                    res1[name] = link
                else:  # it's country file
                    res2[name] = link
            except:  # FIXME: add exception class here.
                name = e.text
                res2[name] = ''

        # create the metadata dataframe
        md_dataset = pd.DataFrame(columns=['name', 'url', 'type'])
        md_dataset['name'] = list(res1.keys())
        md_dataset['url'] = list(res1.values())
        md_dataset['type'] = 'dataset'
        md_country = pd.DataFrame(columns=['name', 'url', 'type'])
        md_country['name'] = list(res2.keys())
        md_country['url'] = list(res2.values())
        md_country['type'] = 'country'

        self.metadata = pd.concat([md_dataset, md_country], ignore_index=True)

        return self.metadata

    def bulk_download(self, out_dir, data_type=None):

        if self.metadata is None:
            self.load_metadata()
        metadata = self.metadata

        if data_type:
            to_download = metadata[metadata['type'] == data_type]
        else:
            to_download = metadata

        for i, row in to_download.iterrows():

            name = row['name']
            path = row['url']

            file_url = urljoin(self.url, path)
            res = requests.get(file_url, stream=True, verify=False)
            fn = osp.join(out_dir, f'{name}.xlsx')

            print("downloading {} to {}".format(file_url, fn))
            with open(fn, 'wb') as f:
                for chunk in res.iter_content(chunk_size=1024):
                    if chunk:
                        f.write(chunk)
                        f.flush()
                f.close()
        print('Done downloading source files.')



# -*- coding: utf-8 -*-

"""main ingredient class"""

__all__ = ['Ingredient', 'ConceptIngredient', 'EntityIngredient', 'DataPointIngredient',
           'ingredient_from_dict', 'key_to_list', 'get_ingredient_class']


import fnmatch
import logging
import os
from collections.abc import Mapping, Sequence
from itertools import product
from functools import singledispatch
from typing import Optional, Union, Dict

import shutil

import numpy as np
import pandas as pd
import dask.dataframe as dd

import attr
from abc import ABC, abstractmethod

from ddf_utils.model.package import DDFcsv
from ddf_utils.model.ddf import DDF
from ddf_utils.model.repo import Repo, is_url
from ddf_utils.str import format_float_digits
from ..exceptions import IngredientError
from ..helpers import gen_sym, query, read_opt, sort_df, read_local_ddf, create_dsk


logger = logging.getLogger('Ingredient')


@attr.s
class Ingredient(ABC):
    """Protocol class for all ingredients.

    all ingredients should have following format:

    .. code-block:: yaml

        id: example-ingredient
        dataset: ddf--example--dataset
        key: "geo,time"  # key columns of ingredient
        value:  # only include concepts listed here
          - concept_1
          - concept_2
        filter:  # select rows by column values
          geo:  # only keep datapoint where `geo` is in [swe, usa, chn]
            - swe
            - usa
            - chn

    The other way to define the ingredient data is using the ``data``
    keyword to include external csv file, or inline the data in the
    ingredient definition. Example:

    .. code-block:: yaml

       id: example-ingredient
       key: concept
       data: external_concepts.csv

    On-the-fly ingredient:

    .. code-block:: yaml

       id: example-ingredient
       key: concept
       data:
           - concept: concept_1
             name: concept_name_1
             concept_type: string
             description: concept_description_1
           - concept: concept_2
             name: concept_name_2
             concept_type: measure
             description: concept_description_2
    """
    # TODO: make some of these attributes _Frozen_
    # currently attrs doesn't support frozen attrs, check issue 133 for attrs
    id: str = attr.ib()
    key: Union[list, str] = attr.ib()
    value: Union[list, dict, str] = attr.ib(default='*')
    dataset: str = attr.ib(default=None)
    data: dict = attr.ib(default=None)
    row_filter: dict = attr.ib(default=None)
    base_dir: str = attr.ib(default='./')

    is_procedure_result: bool = attr.ib(default=False, init=False)  # if data is created by a procedure
    data_computed: dict = attr.ib(default=None, init=False)  # cached result from get_data() or a procedure.

    dtype = 'abc'

    @property
    def ingredient_type(self):
        if self.is_procedure_result:
            return 'procedure_result'
        if self.dataset is not None:
            return 'ddf'
        else:
            if isinstance(self.data, str):
                return 'external'
            else:
                return 'inline'

    @property
    def dataset_path(self):
        """return the full path to ingredient's dataset if the ingredient is from local ddf dataset."""
        if self.ingredient_type == 'ddf':
            if os.path.isabs(self.dataset):
                return self.dataset
            return os.path.join(self.base_dir, self.dataset)
        return None

    @property
    def ddf_id(self):
        if self.ingredient_type == 'ddf':
            ddf = read_local_ddf(self.dataset_path)
            return ddf.props.get('name', self.dataset)
        return None

    @classmethod
    def from_procedure_result(cls, id, key, data_computed: dict):
        res = cls(id, key)
        for _, df in data_computed.items():
            if isinstance(key, list):
                for k in key:
                    assert k in df.columns, "the key {} not in data!".format(k)
            else:
                assert key in df.columns, "the key {} not in data!".format(key)
        res.data_computed = data_computed
        res.is_procedure_result = True
        return res

    @abstractmethod
    def get_data(self):
        ...

    @abstractmethod
    def serve(self, *args, **kwargs):
        """serving data to disk"""
        ...

    @staticmethod
    def filter_row(data: dict, row_filter):
        """return the rows selected by row_filter."""
        if row_filter is None:
            return data
        for k, df in data.items():
            data[k] = query(df, row_filter, available_scopes=df.columns)
        return data


@attr.s
class ConceptIngredient(Ingredient):
    dtype = 'concepts'

    def __attrs_post_init__(self):
        if isinstance(self.key, list):
            if len(self.key) != 1 or self.key[0] != 'concept':
                raise IngredientError('concept ingredient key must be "concept"')
        self.key = 'concept'

    def get_data(self) -> Dict[str, pd.DataFrame]:
        logger.info('evaluating data for {}'.format(self.id))
        if self.data_computed is not None:
            return self.data_computed

        ingredient_type = self.ingredient_type
        if ingredient_type == 'ddf':
            self.data_computed = self.get_data_from_ddf_dataset(self.dataset_path, self.value, self.row_filter)
        if ingredient_type == 'external':
            self.data_computed = self.get_data_from_external_csv(self.data, self.key, self.row_filter)
        if ingredient_type == 'inline':
            self.data_computed = self.get_data_from_inline_data(self.data, self.key, self.row_filter)

        return self.data_computed

    @staticmethod
    def get_data_from_ddf_dataset(dataset_path, value, row_filter):
        ddf = read_local_ddf(dataset_path)
        concepts = ddf.concepts
        df = pd.DataFrame.from_records([x.to_dict() for x in concepts.values()])
        if value != '*':
            if isinstance(value, Mapping):
                assert len(value) == 1
                assert list(value.keys())[0] in ['$in', '$nin']
                kw = list(value.keys())[0]
                if kw == '$in':
                    data = {'concept': df[value[kw]]}
                else:
                    data = {'concept': df[df.columns.drop(value[kw])]}
            else:
                data = {'concept': df[value]}
        else:
            data = {'concept': df}

        data = ConceptIngredient.filter_row(data, row_filter)
        return data

    @staticmethod
    def get_data_from_external_csv(file_path, key, row_filter):
        df = pd.read_csv(file_path)
        data = {key: df}
        data = ConceptIngredient.filter_row(data, row_filter)
        return data

    @staticmethod
    def get_data_from_inline_data(data, key, row_filter):
        df = pd.DataFrame.from_records(data)
        data = {key: df}
        data = ConceptIngredient.filter_row(data, row_filter)
        return data

    def serve(self, outpath, **options):
        if not self.data_computed:
            self.get_data()
        filename = read_opt(options, 'file_name', default='ddf--concepts.csv')
        subpath = read_opt(options, 'path', default=None)
        custom_column_order = read_opt(options, 'custom_column_order', default=None)
        outpath = _handel_subpath(outpath, subpath)
        logger.info('serving: {}'.format(self.id))
        for _, df_ in self.data_computed.items():
            # change boolean into string
            # and remove tailing spaces
            df = df_.copy()
            for i, v in df.dtypes.iteritems():
                if v == 'bool':
                    df[i] = df[i].map(lambda x: str(x).upper())
                if v == 'object':
                    df[i] = df[i].str.strip()
            path = os.path.join(outpath, filename)
            df = sort_df(df, key='concept', custom_column_order=custom_column_order)
            df.to_csv(path, index=False, encoding='utf8')


@attr.s
class EntityIngredient(Ingredient):
    dtype = 'entities'

    def __attrs_post_init__(self):
        if isinstance(self.key, list):
            if len(self.key) != 1:
                raise IngredientError('entity ingredient must have only one primary key')
            self.key = self.key[0]

    def get_data(self) -> Dict[str, pd.DataFrame]:
        logger.info('evaluating data for {}'.format(self.id))
        if self.data_computed is not None:
            return self.data_computed

        ingredient_type = self.ingredient_type
        if ingredient_type == 'ddf':
            self.data_computed = self.get_data_from_ddf_dataset(self.dataset_path, self.key, self.value, self.row_filter)
        if ingredient_type == 'external':
            self.data_computed = self.get_data_from_external_csv(self.data, self.key, self.row_filter)
        if ingredient_type == 'inline':
            self.data_computed = self.get_data_from_inline_data(self.data, self.key, self.row_filter)

        return self.data_computed

    @staticmethod
    def get_data_from_ddf_dataset(dataset_path, key, value, row_filter):
        ddf = read_local_ddf(dataset_path)
        if ddf.concepts[key].concept_type == 'entity_domain':
            domain = key
            eset = None
        else:
            domain = ddf.concepts[key].props['domain']
            eset = key
        df = pd.DataFrame.from_records(ddf.entities[domain].to_dict(eset))
        if value != '*':
            if isinstance(value, Mapping):
                assert len(value) == 1
                assert list(value.keys())[0] in ['$in', '$nin']
                kw = list(value.keys())[0]
                if kw == '$in':
                    cols = [key, *value[kw]]
                    data = {key: df[cols]}
                else:
                    cols = [key, *df.columns.drop(value[kw])]
                    data = {key: df[cols]}
            else:
                cols = [key, *value]
                data = {key: df[cols]}
        else:
            data = {key: df}

        data = EntityIngredient.filter_row(data, row_filter)
        return data

    @staticmethod
    def get_data_from_external_csv(file_path, key, row_filter):
        df = pd.read_csv(file_path)
        data = {key: df}
        data = EntityIngredient.filter_row(data, row_filter)
        return data

    @staticmethod
    def get_data_from_inline_data(data, key, row_filter):
        df = pd.DataFrame.from_records(data)
        data = {key: df}
        data = EntityIngredient.filter_row(data, row_filter)
        return data

    def serve(self, outpath, **options):
        if not self.data_computed:
            self.get_data()
        sets = []
        no_keep_sets = options.get('no_keep_sets', False)  # serve as entity domain
        subpath = options.get('path', None)
        outpath = _handel_subpath(outpath, subpath)
        custom_column_order = options.get('custom_column_order', None)
        logger.info('serving: {}'.format(self.id))
        for k, df in self.data_computed.items():
            # handling is-- headers
            for c in df.columns:
                # changing to upper case is no longer required because it will be upper case string from DDFcsv
                # if df.dtypes[c] == 'bool':  # inferred boolean values
                #     df[c] = df[c].map(lambda x: str(x).upper() if not pd.isnull(x) else x)
                if c.startswith('is--'):  # is-- columns
                    if no_keep_sets:
                        df = df.drop(c, axis=1)
                    else:
                        sets.append(c[4:])
                        # df[c] = df[c].map(lambda x: str(x).upper() if not pd.isnull(x) else x)
            domain = self.key
            if k == domain:
                if len(sets) == 0:
                    path = os.path.join(outpath, 'ddf--entities--{}.csv'.format(k))
                    df = sort_df(df, key=domain, custom_column_order=custom_column_order)
                    df.to_csv(path, index=False, encoding='utf8')
                else:
                    for s in sets:
                        path = os.path.join(outpath, 'ddf--entities--{}--{}.csv'.format(k, s))
                        col = 'is--' + s
                        df_ = df[df[col] == 'TRUE'].dropna(axis=1, how='all')
                        if df_.empty:
                            logger.warning("empty dataframe for %s, not serving", str(s))
                            continue
                        df_ = df_.loc[:, lambda x: ~x.columns.str.startswith('is--')].copy()
                        df_[col] = 'TRUE'
                        df_ = df_.rename({k: s}, axis=1)  # use set name as primary key column name
                        df_ = sort_df(df_, key=s, custom_column_order=custom_column_order)
                        df_.to_csv(path, index=False, encoding='utf8')
                    # serve entities not in any sets
                    is_headers = list(map(lambda x: 'is--' + x, sets))
                    noset = []
                    for i, row in df.iterrows():
                        # import pdb; pdb.set_trace()
                        if (row[is_headers].fillna('FALSE') == 'FALSE').all():
                            noset.append(i)
                    if len(noset) > 0:
                        df_noset = df.loc[noset].drop(is_headers, axis=1).dropna(axis=1, how='all')
                        path = os.path.join(outpath, 'ddf--entities--{}.csv'.format(k))
                        df_noset = sort_df(df_noset, key=domain)
                        df_noset.to_csv(path, index=False)
            else:
                # FIXME: is it even possible that self.key(domain) is not same as k?
                path = os.path.join(outpath, 'ddf--entities--{}--{}.csv'.format(domain, k))
                df = sort_df(df, key=k, custom_column_order=custom_column_order)
                df.to_csv(path, index=False, encoding='utf8')


@attr.s
class DataPointIngredient(Ingredient):
    dtype = 'datapoints'

    def __attrs_post_init__(self):
        if isinstance(self.key, str):
            self.key = key_to_list(self.key)

    def get_data(self) -> Dict[str, dd.DataFrame]:
        logger.info('evaluating data for {}'.format(self.id))
        if self.data_computed is not None:
            return self.data_computed

        ingredient_type = self.ingredient_type
        if ingredient_type == 'ddf':
            self.data_computed = self.get_data_from_ddf_dataset(self.id, self.dataset_path, self.key,
                                                                self.value, self.row_filter)
        if ingredient_type == 'external':
            self.data_computed = self.get_data_from_external_csv(self.data, self.key, self.row_filter)
        if ingredient_type == 'inline':
            self.data_computed = self.get_data_from_inline_data(self.data, self.key, self.row_filter)

        return self.data_computed

    @classmethod
    def from_procedure_result(cls, id, key, data_computed: dict):
        if isinstance(key, str):
            key = key_to_list(key)
        res = cls(id, key)
        for _, df in data_computed.items():
            for k in key:
                assert k in df.columns, "the key {} not in data!".format(k)
        res.data_computed = create_dsk(data_computed)
        res.is_procedure_result = True
        return res

    @staticmethod
    def get_data_from_ddf_dataset(id, dataset_path, key, value, row_filter):
        data = dict()
        ddf = read_local_ddf(dataset_path)
        ddf_id = ddf.props.get('name', dataset_path)

        if value == '*':
            for i in ddf.indicators(by=key):
                data[i] = ddf.get_datapoints(i, by=key).data
        else:
            if isinstance(value, Sequence):  # just a list of indicators to include
                for i in value:
                    if i in ddf.indicators(by=key):
                        data[i] = ddf.get_datapoints(i, by=key).data
                    else:
                        logger.warning("indicator {} not found in dataset {}".format(i, ddf_id))
            else:  # a dictionary with queries
                assert len(value) == 1
                assert list(value.keys())[0] in ['$in', '$nin']
                for keyword, items in value.items():
                    if keyword == '$in':
                        for i in items:
                            matches = fnmatch.filter(ddf.indicators(by=key), i)
                            if len(matches) == 0:
                                logger.warning("indicator matching {} not found in dataset {}".format(i, ddf_id))
                                continue
                            for m in matches:
                                data[m] = ddf.get_datapoints(m, by=key).data
                    else:
                        all_indicators = ddf.indicators(by=key)
                        matches = set(all_indicators) - set(fnmatch.filter(all_indicators, items[0]))
                        for i in items[1:]:
                            matches = matches - set(fnmatch.filter(list(matches), i))
                        if len(matches) == 0:
                            logger.warning("indicators matching the value descriptor not "
                                           "found in dataset " + ddf_id)
                            continue
                        for m in matches:
                            data[m] = ddf.get_datapoints(m, by=key).data
        if len(data) == 0:
            raise IngredientError('no datapoint found for the ingredient: ' + id)

        data = DataPointIngredient.filter_row(data, row_filter)

        return data

    @staticmethod
    def get_data_from_external_csv(file_path, key, row_filter):
        df = pd.read_csv(file_path)
        data = {}
        df = df.set_index(key)
        for c in df.columns:
            data[c] = df[c]
        data = DataPointIngredient.filter_row(data, row_filter)
        data = create_dsk(data, parts=1)
        return data

    @staticmethod
    def get_data_from_inline_data(data, key, row_filter):
        df = pd.DataFrame.from_records(data)
        data = {}
        df = df.set_index(key)
        for c in df.columns:
            data[c] = df[[c]].reset_index()
        data = DataPointIngredient.filter_row(data, row_filter)
        data = create_dsk(data, parts=1)
        return data

    def compute(self) -> Dict[str, pd.DataFrame]:
        """return a pandas dataframe version of self.data"""
        new_data = dict()
        if not self.data_computed:
            self.data_computed = self.get_data()
        for k, df in self.data_computed.items():
            new_data[k] = df.compute()
        return new_data

    def serve(self, outpath, **options):
        digits = read_opt(options, 'digits', default=5)
        dont_serve_empty = read_opt(options, 'drop_empty_datapoints', default=True)
        split_by = read_opt(options, 'split_datapoints_by', default=False)
        subpath = read_opt(options, 'path', default=None)
        custom_key_order = read_opt(options, 'custom_key_order', default=None)
        outpath = _handel_subpath(outpath, subpath)

        logger.info('serving: {}'.format(self.id))

        def to_disk(df_input, k, path, order=None):
            df = df_input.copy()
            if not np.issubdtype(df[k].dtype, np.number):
                try:
                    df[k] = df[k].astype(float)
                    df[k] = df[k].map(lambda x: format_float_digits(x, digits))
                except ValueError:
                    logger.warning("data not numeric: " + k)
            else:
                df[k] = df[k].map(lambda x: format_float_digits(x, digits))
            # sort if custom key order defined
            if order:
                indicator_col, = set(df.columns) - set(order)
                new_cols = [*order, indicator_col]
                df = df[new_cols]
            df.to_csv(path, encoding='utf8', index=False)

        # check if custom key match indicator key
        if custom_key_order:
            if set(custom_key_order) != set(self.key):
                logger.warning("custom keys are not same as ingredient keys")
                logger.warning("ignoring custom_key_order setting")
                custom_key_order = None

        # compute all dask dataframe to pandas dataframe and save to csv files
        # TODO: maybe no need to convert to pandas?
        for k, df in self.compute().items():
            if df.empty:
                logger.info('empty data for {}'.format(k))
                if dont_serve_empty:
                    continue
            by = self.key
            df = sort_df(df, key=by, sort_key_columns=False)
            if not split_by:
                columns = [*by, k]
                path = os.path.join(
                    outpath,
                    'ddf--datapoints--{}--by--{}.csv'.format(k, '--'.join(by)))
                to_disk(df[columns], k, path, custom_key_order)
            else:
                # split datapoints by entities. Firstly we calculate all possible
                # combinations of entities, and then filter the dataframe, create
                # file names and save them into disk.
                values = list()
                [values.append(df[col].unique()) for col in split_by]
                all_combinations = product(*values)

                if len(by) > len(split_by):
                    by = list(set(by) - set(split_by))
                    columns = [*sorted(split_by), *sorted(by), k]
                else:
                    by = None
                    columns = [*sorted(split_by), k]

                for comb in all_combinations:
                    query = ''
                    entity_strings = list()
                    for entity, value in zip(split_by, comb):
                        value_ = str(value)
                        entity_strings.append(entity + '-' + value_)
                        if len(query) > 0:
                            query = query + " and " + "{} == '{}'".format(entity, value_)
                        else:
                            query = "{} == '{}'".format(entity, value_)

                    if by:
                        path = os.path.join(
                            outpath,
                            'ddf--datapoints--{}--by--{}--{}.csv'.format(
                                k,
                                '--'.join(sorted(entity_strings)),
                                '--'.join(sorted(by))
                            )
                        )
                    else:
                        path = os.path.join(
                            outpath,
                            'ddf--datapoints--{}--by--{}.csv'.format(
                                k,
                                '--'.join(sorted(entity_strings))
                            )
                        )
                    # logger.debug('query is: ' + query)
                    df_part = df.query(query)
                    to_disk(df_part[columns], k, path, custom_key_order)


@attr.s
class SynonymIngredient(Ingredient):
    dtype = 'synonyms'

    def __attrs_post_init__(self):
        if isinstance(self.key, str):
            self.key = key_to_list(self.key)

    def get_data(self):
        logger.info('evaluating data for {}'.format(self.id))
        if self.data_computed is not None:
            return self.data_computed

        ingredient_type = self.ingredient_type
        if ingredient_type == 'ddf':
            self.data_computed = self.get_data_from_ddf_dataset(self.dataset_path, self.key)
        if ingredient_type == 'external':
            self.data_computed = self.get_data_from_external_csv(self.data, self.key)
        if ingredient_type == 'inline':
            self.data_computed = self.get_data_from_inline_data(self.data, self.key)

        return self.data_computed

    @staticmethod
    def get_data_from_ddf_dataset(dataset_path, key):
        ddf = read_local_ddf(dataset_path)
        key.remove('synonym')
        k = key[0]
        synonyms = ddf.synonyms[k].synonyms  # this is a dictionary, not dataframe.
        # nothing to do with `value` and `row_filter`.
        data = {k: synonyms}
        return data

    @staticmethod
    def get_data_from_external_csv(file_path, key):
        df = pd.read_csv(file_path)
        key.remove('synonym')
        k = key[0]
        data = {k: df.set_index('synonym')[k].to_dict()}
        # nothing to do with `value` and `row_filter`.
        return data

    @staticmethod
    def get_data_from_inline_data(data, key):
        df = pd.DataFrame.from_records(data)
        key.remove('synonym')
        k = key[0]
        data = {k: df.set_index('synonym')[k].to_dict()}
        # nothing to do with `value` and `row_filter`.
        return data

    def serve(self, *args, **kwargs):
        raise NotImplementedError


# functions for creating Ingredient from recipe definitions

def key_to_list(key):
    """make a list that contains primaryKey of this ingredient"""
    return [x.strip() for x in key.split(',')]


def infer_type_from_keys(keys: list):
    """infer ddf data type from the primary key"""
    if len(keys) == 1:
        if keys[0] == 'concept':
            return 'concepts'
        else:
            return 'entities'
    if 'synonym' in keys:
        return 'synonyms'
    else:
        return 'datapoints'


def ingredient_from_dict(dictionary: dict, **chef_options) -> Ingredient:
    """create ingredient from recipe definition and options. Parameters
    for ingredient should be passed in a dictionary. See the doc for
    :ref:`ingredient def` or :py:class:`ddf_utils.chef.model.ingredient.Ingredient` for
    available parameters.
    """
    ingred_id = read_opt(dictionary, 'id', default=None, method='pop')
    dataset = read_opt(dictionary, 'dataset', default=None, method='pop')
    data = read_opt(dictionary, 'data', default=None, method='pop')
    key = read_opt(dictionary, 'key', required=True, method='pop')
    value = read_opt(dictionary, 'value', default='*', method='pop')
    row_filter = read_opt(dictionary, 'filter', default=None, method='pop')

    dataset_dir = read_opt(chef_options, 'ddf_dir', default='./')
    external_csv_dir = read_opt(chef_options, 'external_csv_dir', default='./')

    if ((dataset is None and data is None) or
            (dataset is not None and data is not None)):
        raise ValueError("Please provide either `dataset` or `data` for {}".format(ingred_id))

    if ingred_id is None:
        ingred_id = gen_sym('inline', None, dictionary)

    if dataset is not None:
        ingredient_type = 'ddf'
    else:
        if isinstance(data, str):
            ingredient_type = 'external'
        else:
            ingredient_type = 'inline'

    if len(dictionary.keys()) > 0:
        logger.warning("Ignoring following keys: {}".format(list(dictionary.keys())))

    if ingredient_type == 'ddf':
        if is_url(dataset):  # data will read from github dataset
            repo = Repo(dataset, base_path=dataset_dir)  # this will clone the repo to dataset_dir if it doesn't exist
            dataset = os.path.relpath(repo.local_path, dataset_dir)
        # else:
        #     dataset = os.path.join(dataset_dir, dataset)

    if ingredient_type == 'external':
        data = os.path.join(external_csv_dir, data)

    keys = key_to_list(key)
    dtype = infer_type_from_keys(keys)
    if dtype == 'concepts':
        return ConceptIngredient(id=ingred_id, dataset=dataset, data=data, key=key, value=value,
                                 row_filter=row_filter, base_dir=dataset_dir)
    if dtype == 'entities':
        return EntityIngredient(id=ingred_id, dataset=dataset, data=data, key=key, value=value,
                                row_filter=row_filter, base_dir=dataset_dir)
    if dtype == 'datapoints':
        return DataPointIngredient(id=ingred_id, dataset=dataset, data=data, key=key, value=value,
                                   row_filter=row_filter, base_dir=dataset_dir)
    if dtype == 'synonyms':
        return SynonymIngredient(id=ingred_id, dataset=dataset, data=data, key=key, value=value,
                                 row_filter=row_filter, base_dir=dataset_dir)


def get_ingredient_class(cls):
    d = {
        'concepts': ConceptIngredient,
        'entities': EntityIngredient,
        'datapoints': DataPointIngredient,
        'synonyms': SynonymIngredient
    }
    return d[cls]


def _handel_subpath(outpath, subpath):
    if subpath is None:
        newpath = outpath
    else:
        newpath = os.path.join(outpath, subpath)

    os.makedirs(newpath, exist_ok=True)

    return newpath



# -*- coding: utf-8 -*-

"""Functions to load data from cdiac co2 website

source link: The `CDIAC ftp site`_

.. _`CDIAC ftp site`: http://cdiac.ess-dive.lbl.gov/ftp/ndp030/CSV-FILES/
"""

import os.path as osp

import requests
import pandas as pd

from urllib.parse import urljoin

from . common import DataFactory


class CDIACLoader(DataFactory):

    url = 'http://cdiac.ess-dive.lbl.gov/ftp/ndp030/CSV-FILES/'

    def load_metadata(self):
        """load the file info listed in the CDIAC website."""
        r = requests.get(self.url)
        assert r.status_code == 200, 'load cdiac website failed!'

        data = pd.read_html(r.content)[0]

        data = data.dropna(axis=1, how='all').dropna(axis=0, how='all')
        data = data.iloc[2:]

        data['Last modified'] = pd.to_datetime(data['Last modified'])

        self.metadata = data
        return data

    def has_newer_source(self, date):
        if self.metadata is None:
            self.load_metadata()
        metadata = self.metadata
        newer = metadata[metadata['Last modified'] > date].values
        if len(newer) > 0:
            return True
        return False

    def bulk_download(self, out_dir):
        """download the latest nation/global file"""
        if self.metadata is None:
            self.load_metadata()
        metadata = self.metadata

        # latest nation/global file names
        lm = 'Last modified'
        nation_fn = (metadata[metadata['Name'].str.startswith('nation')]
                     .sort_values(by=[lm, 'Name'], ascending=False)['Name']
                     .values[0])
        global_fn = (metadata[metadata['Name'].str.startswith('global')]
                     .sort_values(by=[lm, 'Name'], ascending=False)['Name']
                     .values[0])

        # download files
        nation_url = urljoin(self.url, nation_fn)
        nation_fs_path = osp.join(out_dir, 'nation.csv')

        with open(osp.expanduser(nation_fs_path), 'wb') as f:
            r = requests.get(nation_url)
            f.write(r.content)
            f.close()

        global_url = urljoin(self.url, global_fn)
        global_fs_path = osp.join(out_dir, 'global.csv')

        with open(osp.expanduser(global_fs_path), 'wb') as f:
            r = requests.get(global_url)
            f.write(r.content)
            f.close()

        return out_dir



# -*- coding: utf-8 -*-

"""merge procedure for recipes"""

import logging
import time

from typing import List, Dict, Union

import numpy as np
import pandas as pd
import dask.dataframe as dd

from .. exceptions import ProcedureError
from .. helpers import debuggable
from .. model.ingredient import Ingredient, get_ingredient_class
from .. model.chef import Chef


logger = logging.getLogger('merge')


@debuggable
def merge(chef: Chef, ingredients: List[Ingredient], result, deep=False) -> Ingredient:
    """merge a list of ingredients

    The ingredients will be merged one by one in the order of how they are provided to this
    function. Later ones will overwrite the previous merged results.

    Procedure format:

    .. code-block:: yaml

       procedure: merge
       ingredients:  # list of ingredient id
         - ingredient_id_1
         - ingredient_id_2
         - ingredient_id_3
         # ...
       result: str  # new ingredient id
       options:
         deep: bool  # use deep merge if true

    Parameters
    ----------
    chef: Chef
        a Chef instance
    ingredients:
        Any numbers of ingredients to be merged

    Keyword Args
    ------------
    deep: `bool`, optional
        if True, then do deep merging. Default is False

    Notes
    -----
    **deep merge** is when we check every datapoint for existence
    if false, overwrite is on the file level. If key-value
    (e.g. geo,year-population_total) exists, whole file gets overwritten
    if true, overwrite is on the row level. If values
    (e.g. afr,2015-population_total) exists, it gets overwritten, if it doesn't it stays
    """
    # ingredients = [chef.dag.get_node(x).evaluate() for x in ingredients]
    logger.info("merge: " + str([i.id for i in ingredients]))

    # assert that dtype and key are same in all dataframe
    try:
        for x in ingredients[1:]:
            assert set(x.key) == set(ingredients[0].key)
        assert len(set([x.dtype for x in ingredients])) == 1
    except (AssertionError, TypeError):
        log1 = "multiple dtype/key detected: \n"
        log2 = "\n".join(["{}: {}, {}".format(x.id, x.dtype, x.key) for x in ingredients])
        logger.warning(log1 + log2)
        raise ProcedureError("can't merge data with multiple dtype/key!")

    # get the dtype and index
    # we have assert dtype and key is unique, so we take it from
    # the first ingredient
    dtype = ingredients[0].dtype

    if dtype == 'datapoints':
        index_col = ingredients[0].key
        newkey = ','.join(index_col)
    else:
        index_col = ingredients[0].key
        newkey = index_col

    if deep:
        logger.info("merge: doing deep merge")
    # merge data from ingredients one by one.
    res_all = {}

    for i in ingredients:
        res_all = _merge_two(res_all, i.get_data(), index_col, dtype, deep)

    if not result:
        result = 'all_data_merged_' + str(int(time.time() * 1000))

    return get_ingredient_class(dtype).from_procedure_result(result, newkey, data_computed=res_all)


def __get_last_item(ser):
    """get the last valid item of a Series, or Nan."""
    ser_ = ser.dropna()
    if ser_.last_valid_index() is None:
        return np.nan
    else:
        return ser_.values[-1]


def _merge_two(left: Dict[str, Union[pd.DataFrame, dd.DataFrame]],
               right: Dict[str, Union[pd.DataFrame, dd.DataFrame]],
               index_col: Union[List, str],
               dtype: str, deep=False) -> Dict[str, pd.DataFrame]:
    """merge 2 ingredient data."""
    if len(left) == 0:
        return right

    res_data = {}

    # for datapoints we use dask to help performance.
    if dtype == 'datapoints':
        res_data = dict([(k, v) for k, v in left.items()])
        if deep:
            for k, df in right.items():
                if k in left.keys():
                    columns = left[k].columns.values
                    # res_data[k] = left[k].append(df[columns], interleave_partitions=True)
                    res_data[k] = dd.concat([left[k], df[columns]], axis=0, interleave_partitions=True)
                    res_data[k] = res_data[k].drop_duplicates(subset=index_col, keep='last')
                    # res_data[k] = res_data[k].sort_values(by=index_col)
                else:
                    res_data[k] = df
        else:
            for k, df in right.items():
                res_data[k] = df

    # for concepts/entities, we don't need to use dask.
    elif dtype == 'concepts':

        left_df = pd.concat([x for x in left.values()], sort=False)
        right_df = pd.concat([x for x in right.values()], sort=False)

        if deep:
            merged = left_df.append(right_df, sort=False)
            res = merged.groupby(by=index_col).agg(__get_last_item)
            res_data = {'concept': res.reset_index()}
        else:
            res_data = {'concept':
                        right_df.drop_duplicates(subset='concept',
                                                 keep='last')}
        res_data = res_data

    else:  # entities
        if deep:
            for k, df in right.items():
                if k in left.keys():
                    left[k] = left[k].append(df, ignore_index=True, sort=False)
                    left[k] = left[k].groupby(index_col).agg(__get_last_item).reset_index()
                else:
                    left[k] = df
        else:
            for k, df in right.items():
                left[k] = df
        res_data = left

    return res_data



# -*- coding: utf-8 -*-

"""all procedures for recipes"""

import logging
from typing import List

import pandas as pd

from .. helpers import debuggable
from .. model.ingredient import DataPointIngredient
from .. model.chef import Chef

logger = logging.getLogger('trend_bridge')


@debuggable
def trend_bridge(chef: Chef, ingredients: List[DataPointIngredient], bridge_start, bridge_end,
                 bridge_length, bridge_on, result, target_column=None) -> DataPointIngredient:
    """run trend bridge on ingredients

    .. highlight:: yaml

    Procedure format:

    ::

      procedure: trend_bridge
      ingredients:
        - data_ingredient                 # optional, if not set defaults to empty ingredient
      result: data_bridged
      options:
        bridge_start:
            ingredient: old_data_ingredient # optional, if not set then assume it's the input ingredient
            column:
              - concept_old_data
        bridge_end:
            ingredient: new_data_ingredient # optional, if not set then assume it's the input ingredient
            column:
              - concept_new_data
        bridge_length: 5                  # steps in time. If year, years, if days, days.
        bridge_on: time                   # the index column to build the bridge with
        target_column:
              - concept_in_result  # overwrites if exists. creates if not exists. default to bridge_end.column

    Parameters
    ----------
    chef: Chef
        A Chef instance
    ingredients : list
        The input ingredient. The bridged result will be merged in to this ingredient. If this is
        None, then the only the bridged result will be returned
    bridge_start : dict
        Describe the start of bridge
    bridge_end : dict
        Describe the end of bridge
    bridge_length : int
        The size of bridge
    bridge_on : `str`
        The column to bridge
    result : `str`
        The output ingredient id

    Keyword Args
    ------------
    target_column : `list`, optional
        The column name of the bridge result. default to `bridge_end.column`

    See Also
    --------
    :py:func:`ddf_utils.transformer.trend_bridge` : related function in transformer module
    """
    from ... transformer import trend_bridge as tb

    # check parameters
    if ingredients is None or ingredients == []:
        assert 'ingredient' in bridge_start.keys()
        assert 'ingredient' in bridge_end.keys()
        ingredient = None
    else:
        assert len(ingredients) <= 1, "procedure only support 1 ingredient for now."
        # ingredient = chef.dag.get_node(ingredients[0]).evaluate()
        ingredient = ingredients[0]

    # get data for start and end
    if 'ingredient' in bridge_start.keys():
        start = chef.dag.get_node(bridge_start['ingredient']).evaluate()
    else:
        start = ingredient
    if 'ingredient' in bridge_end.keys():
        end = chef.dag.get_node(bridge_end['ingredient']).evaluate()
    else:
        end = ingredient

    assert start.dtype == 'datapoints'
    assert end.dtype == 'datapoints'

    assert len(bridge_end['column']) == len(bridge_start['column']),\
        "columns length in bridge_start and bridge_end should be the same!"

    logger.info("trend_bridge: {} and {}".format(start.id, end.id))

    if target_column is None:
        target_column = bridge_end['column']

    # get the column to group. Because datapoints are multidimensional, but we only
    # bridge them in one column, so we should group other columns.
    try:
        assert set(start.key) == set(end.key)
    except AssertionError:
        logger.critical("start and end have different keys! {} and {}".format(start.key, end.key))
        raise

    keys = start.key.copy()
    keys.remove(bridge_on)

    # calculate for each column
    new_data = dict()

    start_computed = start.compute()
    end_computed = end.compute()

    for c1, c2, c3 in zip(bridge_start['column'], bridge_end['column'], target_column):
        logger.info("bridge_start: {}, bridge_end: {}, target_column: {}".format(c1, c2, c3))
        start_group = start_computed[c1].set_index(bridge_on).groupby(keys)
        end_group = end_computed[c2].set_index(bridge_on).groupby(keys)

        # get all groups
        # NOTE: currently pandas (0.24) always emits all values from a category when doing groupby, regardless whether
        # the value is actually in the column.
        g1 = list(start_group.groups.keys())
        g2 = list(end_group.groups.keys())
        all_groups = g1.copy()
        for g in g2:
            if g not in all_groups:
                all_groups.append(g)

        # calculate trend bridge on each group
        res_grouped = []
        for g in all_groups:
            try:
                gstart = start_group.get_group(g)[c1].copy()
            except KeyError:
                logger.warning("no data for bridge start: " + str(g))
                gstart = None
            try:
                gend = end_group.get_group(g)[c2].copy()
            except KeyError:
                logger.warning("no data for bridge end: " + str(g))
                gend = None

            if gstart is None and gend is None:
                continue
            if gstart is None:
                bridged = gend
            elif gend is None:
                bridged = gstart
            else:
                bridged = tb(gstart, gend, bridge_length)

            res_grouped.append((g, bridged))

        # combine groups to dataframe
        res = []
        for g, v in res_grouped:
            v.name = c3
            v = v.reset_index()
            if len(keys) == 1:
                assert isinstance(g, str)
                v[keys[0]] = g
            else:
                assert isinstance(g, tuple)
                for i, k in enumerate(keys):
                    v[k] = g[i]
            res.append(v)
        new_data[c3] = pd.concat(res, ignore_index=True)

    # merge in to ingredient in `ingredients` parameter if needed
    if ingredient is not None:
        from . merge import _merge_two
        merged = _merge_two(ingredient.compute(), new_data,
                            start.key, 'datapoints', deep=False)
        return DataPointIngredient.from_procedure_result(result, start.key, merged)
    else:
        return DataPointIngredient.from_procedure_result(result, start.key, new_data)



# -*- coding: utf-8 -*-

"""Functions for IHME

The `GBD result tool`_ at IHME contains all data for GBD results, but
they don't have an open API to query the data. However the website
uses a json endpoint and it doesn't need authorization. So we also
make use of it.

.. _`GBD result tool`: http://ghdx.healthdata.org/gbd-results-tool

"""

import os
import os.path as osp
import math
from time import sleep
import requests
import pandas as pd

from tqdm import tqdm

from ddf_utils.chef.helpers import read_opt
from . common import requests_retry_session, DataFactory, retry


# TODO: add missing context/base configures.
class IHMELoader(DataFactory):
    url_hir = 'http://ghdx.healthdata.org/sites/all/modules/custom/ihme_query_tool/gbd-search/php/hierarchy/'
    url_metadata = 'http://ghdx.healthdata.org/sites/all/modules/custom/ihme_query_tool/gbd-search/php/metadata/'
    url_version = 'http://ghdx.healthdata.org/sites/all/modules/custom/ihme_query_tool/gbd-search/php/version/'
    # url for query data:
    # http://ghdx.healthdata.org/sites/all/modules/custom/ihme_query_tool/gbd-search/php/data.php
    # below is url for download data as zip
    url_data = 'http://ghdx.healthdata.org/sites/all/modules/custom/ihme_query_tool/gbd-search/php/download.php'
    url_task = 'https://s3.healthdata.org/gbd-api-2017-public/{hash}'  # access to download link

    def load_metadata(self):
        """load all codes used in GBD in a dictionary."""

        session = requests_retry_session()
        meta = session.get(self.url_metadata).json()
        versions = session.get(self.url_version).json()

        metadata = {}

        for k in meta['data'].keys():
            metadata[k] = pd.DataFrame.from_dict(meta['data'][k], orient='index')

        metadata['version'] = pd.DataFrame.from_dict(versions['data'], orient='index')
        self.metadata = metadata
        return metadata

    def has_newer_source(self, ver):
        if not self.metadata:
            self.load_metadata()
        metadata = self.metadata
        versions = metadata['version']
        newer = versions[versions['id'] > ver].values
        return bool(len(newer) > 0)

    def download_links(self, url):
        session = requests_retry_session()

        sent_urls = []

        success_results = [
            'Your search returned no results.',
            'success'
        ]  # other results means still working

        while True:
            res_json = session.get(url).json()

            dus = res_json['urls']
            for du in dus:
                if du in sent_urls:
                    continue
                else:
                    sent_urls.append(du)
                    print(du)
                    yield du

            if res_json['state'] in success_results:
                break
            sleep(10)

    def bulk_download(self, out_dir, version, context=None, query=None, **kwargs):
        """download the selected contexts/queries from GBD result tools.

        Either context or query should be supplied. If both are supplied,
        query will be used.

        `context` should be a list of string and `query` should be a list
        of dictionaries containing post requests data.
        """
        if not self.metadata:
            self.load_metadata()

        metadata = self.metadata

        if query is None and context is None:
            raise ValueError('one of context and query should be supplied!')
        elif query is None:
            if isinstance(context, list):
                query = [self._make_query(c, version, **kwargs) for c in context]
            else:
                query = [self._make_query(context, version, **kwargs)]
        else:
            if not isinstance(query, list):
                query = [query]

        taskIDs = set()

        # make a series of queries, the server will response a series of task ids.
        session = requests_retry_session()
        for q in query:
            res_data = session.post(self.url_data, data=q)
            # print(res_data.json())
            if res_data.status_code not in [200, 202]:
                print(res_data.text)
                raise ValueError("status code not 200: {}".format(res_data.status_code))
            if isinstance(res_data.json()['taskID'], list):
                for taskID in res_data.json()['taskID']:
                    taskIDs.add(taskID)
            else:
                taskIDs.add(res_data.json()['taskID'])

        # then, we check each task, download all files linked to the task.
        if len(taskIDs) == 0:
            print('no available results')
            return

        for i in taskIDs:
            url = self.url_task.format(hash=i)
            print('working on {}'.format(url))
            print('check status as http://ghdx.healthdata.org/gbd-results-tool/result/{}'.format(i))
            print('available downloads:')

            for u in self.download_links(url):
                tries = 1
                while tries <= 5:
                    try:
                        self._run_download(u, out_dir, taskID=i)
                        break
                    except (ValueError, requests.exceptions.ConnectionError) as e:
                        if tries == 5:
                            raise
                        print("download interrupted, retrying...")
                        tries = tries + 1

        return [i[:8] for i in taskIDs]

    @retry(times=3)
    def _run_download(self, u, out_dir, taskID):
        '''accept an URL and download it to out_dir'''
        download_file = requests_retry_session().get(u, stream=True, timeout=60)

        if download_file.status_code != 200:
            print(f'can not download source file: {u}')
            return

        if not osp.exists(osp.join(out_dir, taskID[:8])):
            os.mkdir(osp.join(out_dir, taskID[:8]))

        fn = osp.join(out_dir, taskID[:8], osp.basename(u))
        print('downloading {} to {}'.format(u, fn))

        block_size = 1024
        total_size = int(download_file.headers.get('content-length', 0))
        wrote = 0
        with open(fn, 'wb') as f:
            for c in tqdm(download_file.iter_content(chunk_size=block_size),
                          total=math.ceil(total_size // block_size), unit='KB', unit_scale=True):
                f.write(c)
                wrote = wrote + len(c)
                f.flush()
        if wrote != total_size:
            raise ValueError("download failed.")

    def _make_query(self, context, version, **kwargs):
        # metadata
        if not self.metadata:
            self.load_metadata()

        metadata = self.metadata
        ages = metadata['age']['id'].values
        # location: there is a `custom` location. don't include that one.
        locations_md = metadata['location']
        locations = locations_md[locations_md['location_id'] != 'custom']['id'].tolist()
        sexs = metadata['sex']['id'].tolist()
        years = metadata['year']['id'].tolist()
        metrics = metadata['metric']['id'].tolist()
        measures = metadata['measure']['id'].tolist()
        causes = metadata['cause']['id'].tolist()
        # risk/etiology/impairment
        # There are actually 4 types data in this dictionary:
        # risk, etiology, impairment and injury n-codes.
        # however injury n-codes is not enabled.
        # we might need to take of it later.
        rei = metadata['rei']
        # risks = rei[rei['type'] == 'risk']['rei_id'].values
        # etiologys = rei[rei['type'] == 'etiology']['rei_id'].values
        # impairments = rei[rei['type'] == 'impairment']['rei_id'].values

        # others metadata filed not include:
        # - groups
        # - year_range

        queries = {}

        # create query base on context and user input
        if context == 'le':
            measure = read_opt(kwargs, 'measure', default=26)
            metric = read_opt(kwargs, 'metric', default=5)
            cause = read_opt(kwargs, 'cause', default=causes)
            queries.update({
                'measure[]': measure,
                'metric[]': metric,
                'cause[]': cause
            })
        elif context == 'cause':
            measure = read_opt(kwargs, 'measure', default=measures)
            metric = read_opt(kwargs, 'metric', default=[1, 2, 3])
            cause = read_opt(kwargs, 'cause', default=causes)
            queries.update({
                'measure[]': measure,
                'metric[]': metric,
                'cause[]': cause
            })
        elif context in ['risk', 'etiology', 'impairment']:
            # TODO: should be split, don't combine these context here.
            measure = read_opt(kwargs, 'measure', default=measures)
            metric = read_opt(kwargs, 'metric', default=[1, 2, 3])
            context_values = rei[rei['type'] == context]['rei_id'].tolist()
            cause = read_opt(kwargs, 'cause', default=causes)
            queries.update({
                'measure[]': measure,
                'metric[]': metric,
                context+'[]':context_values,
                'rei[]': context_values,
                'cause[]':  cause
            })
        else:
            # SEV/HALE/haqi
            print('not supported context.')
            raise NotImplementedError

        # insert context and version and other configs
        rows = read_opt(kwargs, 'rows', default=10000000)  # the maximum records we can get
        # ^ Note: user guide[1] says it's 500000 row. But actually we can set this to 10000000
        # [1]: http://www.healthdata.org/sites/default/files/files/Data_viz/GBD_2017_Tools_Overview.pdf
        email = read_opt(kwargs, 'email', default='downloader@gapminder.org')
        idsOrNames = read_opt(kwargs, 'idsOrNames', default='ids')           # ids / names / both
        singleOrMult = read_opt(kwargs, 'singleOrMult', default='multiple')  # single / multiple
        base = read_opt(kwargs, 'base', default='single')

        location = read_opt(kwargs, 'location', default=locations)
        age = read_opt(kwargs, 'age', default=ages)
        sex = read_opt(kwargs, 'sex', default=sexs)
        year = read_opt(kwargs, 'year', default=years)

        queries.setdefault('context', context)
        queries.setdefault('version', version)
        queries.setdefault('rows', rows)
        queries.setdefault('email', email)
        queries.setdefault('idsOrNames', idsOrNames)
        queries.setdefault('singleOrMult', singleOrMult)
        queries.setdefault('base', base)
        queries.setdefault('location[]', location)
        queries.setdefault('age[]', age)
        queries.setdefault('sex[]', sex)
        queries.setdefault('year[]', year)

        return queries



# -*- coding: utf-8 -*-

"""extract_concepts procedure for recipes"""

import logging
import numpy as np
import pandas as pd

from typing import List

from .. helpers import debuggable
from .. model.ingredient import Ingredient, ConceptIngredient
from .. model.chef import Chef


logger = logging.getLogger('extract_concepts')


@debuggable
def extract_concepts(chef: Chef, ingredients: List[Ingredient], result,
                     join=None, overwrite=None, include_keys=False) -> ConceptIngredient:
    """extract concepts from other ingredients.

    .. highlight:: yaml

    Procedure format:

    ::

       procedure: extract_concepts
       ingredients:  # list of ingredient id
         - ingredient_id_1
         - ingredient_id_2
       result: str  # new ingredient id
       options:
         join:  # optional
           base: str  # base concept ingredient id
           type: {'full_outer', 'ingredients_outer'}  # default is full_outer
         overwrite:  # overwrite some concept types
           country: entity_set
           year: time
         include_keys: true  # if we should include the primaryKeys concepts

    Parameters
    ----------
    ingredients
        any numbers of ingredient that needs to extract concepts from

    Keyword Args
    ------------
    join : dict, optional
        the base ingredient to join
    overwrite : dict, optional
        overwrite concept types for some concepts
    include_keys : bool, optional
        if we shuld include the primaryKeys of the ingredients, default to false

    See Also
    --------
    :py:func:`ddf_utils.transformer.extract_concepts` : related function in transformer
    module

    Note
    ----
    - all concepts in ingredients in the ``ingredients`` parameter will be extracted
      to a new concept ingredient
    - ``join`` option is optional; if present then the ``base`` will merge with concepts
      from ``ingredients``
    - ``full_outer`` join means get the union of concepts; ``ingredients_outer`` means
      only keep concepts from ``ingredients``

    """

    # ingredients = [chef.dag.get_node(x).evaluate() for x in ingredients]
    logger.info("extract concepts: {}".format([x.id for x in ingredients]))

    if join:
        base = chef.dag.get_node(join['base']).evaluate()
        try:
            join_type = join['type']
        except KeyError:
            join_type = 'full_outer'
        concepts = base.get_data()['concept'].set_index('concept')
    else:
        concepts = pd.DataFrame([], columns=['concept', 'concept_type']).set_index('concept')
        join_type = 'full_outer'

    new_concepts = set()

    for i in ingredients:
        data = i.get_data()
        if i.dtype in ['concepts', 'entities']:
            pks = [i.key]
        else:
            pks = i.key

        for k, df in data.items():
            if include_keys:
                cols = df.columns
            else:
                cols = [x for x in df.columns if x not in pks]

            cat_cols = df.select_dtypes(include=['category']).columns
            for col in cols:
                if col.startswith('is--'):
                    continue
                new_concepts.add(col)
                if col in concepts.index:
                    continue
                # if df.dtypes[col] == 'category':  # doesn't work
                if col in cat_cols:
                    concepts.loc[col, 'concept_type'] = 'string'
                else:
                    concepts.loc[col, 'concept_type'] = 'measure'

    if join_type == 'ingredients_outer':
        # ingredients_outer join: only keep concepts appears in ingredients
        concepts = concepts.loc[new_concepts]

    # add name column if there isn't one
    if 'name' not in concepts.columns:
        concepts['name'] = np.nan
    if 'name' not in concepts.index.values:
        concepts.loc['name', 'concept_type'] = 'string'
        concepts.loc['name', 'name'] = 'Name'
    concepts['name'] = concepts['name'].fillna(
        concepts.index.to_series().map(lambda x: str(x).replace('_', ' ').title()))

    # overwrite some of the types
    if overwrite:
        for k, v in overwrite.items():
            concepts.loc[k, 'concept_type'] = v
    if not result:
        result = 'concepts_extracted'
    return ConceptIngredient.from_procedure_result(result, 'concept',
                                                   data_computed={'concept': concepts.reset_index()})



# -*- coding: utf-8 -*-

"""flatten procedure for recipes"""

import fnmatch
import logging

from typing import List

from .. helpers import debuggable, read_opt
from .. model.ingredient import DataPointIngredient
from .. model.chef import Chef


logger = logging.getLogger('flatten')


@debuggable
def flatten(chef: Chef, ingredients: List[DataPointIngredient], result, **options) -> DataPointIngredient:
    """flattening some dimensions, create new indicators.

    procedure format:

    .. code-block:: yaml

       procedure: flatten
       ingredients:
           - ingredient_to_run
       options:
           flatten_dimensions:
               - entity_1
               - entity_2
           dictionary:
               "concept_name_wildcard": "new_concept_name_template"
           skip_totals_among_entities:
               - entity_1
               - entity_2

    The ``dictionary`` can have multiple entries, for each entry the concepts that matches the key in wildcard
    matching will be flatten to the value, which should be a template string. The variables for the templates
    will be provided with a dictionary contains ``concept``, and all columns from ``flatten_dimensions`` as keys.

    Parameters
    ----------
    chef : Chef
        the Chef instance
    ingredients : list
        a list of ingredients
    result : `str`
        id of result ingredient
    skip_totals_among_entities : list
        a list of total among entities, which we don't add to new indicator names

    Keyword Args
    ------------
    flatten_dimensions: list
        a list of dimension to be flattened
    dictionary: dict
        the dictionary for old name -> new name mapping
    """
    assert len(ingredients) == 1, "procedure only support 1 ingredient for now."

    # ingredient = chef.dag.get_node(ingredients[0]).evaluate()
    ingredient = ingredients[0]
    data = ingredient.get_data()

    logger.info("flatten: " + ingredient.id)

    flatten_dimensions = read_opt(options, 'flatten_dimensions', required=True)
    if not isinstance(flatten_dimensions, list):
        flatten_dimensions = [flatten_dimensions]
    dictionary = read_opt(options, 'dictionary', required=True)
    skip_totals_among_entities = read_opt(options, 'skip_totals_among_entities')

    newkey = [x for x in ingredient.key if x not in flatten_dimensions]
    newkey = ','.join(newkey)

    res = {}
    for from_name_tmpl, new_name_tmpl in dictionary.items():
        dfs = dict([(x, data[x]) for x in fnmatch.filter(data.keys(), from_name_tmpl)])
        for from_name, df_ in dfs.items():
            df = df_.compute()
            grouper = df.groupby(flatten_dimensions)
            for g, _ in grouper.groups.items():
                # logger.warn(g)
                # FIXME: There is an issue for pandas grouper for categorical data
                # where it will return all categories even if it's already filtered
                # it's WIP and refer to pull request #20583 for pandas.
                try:
                    df_ = grouper.get_group(g)
                except KeyError:
                    continue
                if df_.empty:
                    continue
                if not isinstance(g, tuple):
                    g = [g]
                tmpl_dict = dict(zip(flatten_dimensions, g))
                tmpl_dict['concept'] = from_name
                new_name = new_name_tmpl.format(**tmpl_dict)
                # remove totals among entities from name
                if skip_totals_among_entities is not None:
                    for e in skip_totals_among_entities:
                        new_name = new_name.replace('_' + e, '')
                    logger.info('new name w/o total among entities is {}'.format(new_name))
                if new_name in res.keys():
                    # raise ProcedureError("{} already created! check your name template please.".format(new_name))
                    logger.warning("{} already exists! It will be overwritten.".format(new_name))
                res[new_name] = df_.rename(columns={from_name: new_name}).drop(flatten_dimensions, axis=1)

    return DataPointIngredient.from_procedure_result(result, newkey, data_computed=res)
